# -*- coding: utf-8 -*-

"""

  core injection
  ~~~~~~~~~~~~~~

  tools for dependency injection - essentially, walking the
  structure exposed by classes in :py:mod:`core.meta` to
  combine dependencies into compound classes.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# canteen meta
from .meta import Proxy


class Delegate(object):

  """ Delegates property access for a given context to the system DI (dependency
      injection) pool. Items that mix a ``Delegate`` subclass into their MRO
      work transparently with injection. """

  __target__, __bridge__ = (
      None,  # holds injection target for this delegate
      None)  # holds bridge for current class to collapsed component set

  class __metaclass__(type):

    """ Metaclass to prepare new subclasses, wrapped as customized children of
        ``Delegate`` that contain their owner class for DI context. """

    def __new__(mcs, name_or_target, bases=None, properties=None):

      """ Construct a new ``Delegate`` subclass.

          :param name_or_target: Either the ``str`` name of a subtype extending
            ``Delegate`` directly or a name to wrap into an ephemeral
            ``Delegate`` type (mainly for use in labelling).

          :param bases: Unset (``None``) unless extending ``Delegate``,
            otherwise a tuple of ``bases``.

          :param properties: Unset (``None``) unless extending ``Delegate``,
            otherwise a ``dict`` mapping of class-level attributes and values.

          :returns: Factoried ``Delegate`` subtype, wrapped with an injection
            responder closure suitable for use in MRO-based dependency
            resolution. """

      # if it only comes through with a target, it's a subdelegate
      if bases or properties:

        # it's `Delegate` probably - construct it as normal
        name, target = name_or_target, None
        return type.__new__(mcs, name, bases, properties)

      # otherwise, construct with an MRO attribute injection
      name, target = None, name_or_target

      def injection_responder(klass, key):

        """ Injected responder for attribute accesses that hit the DI pool.
            Resolves a ``key`` for a given ``klass``, if possible.

            :param klass: Origin class for which we should generate a DI pool
              on-the-fly. This could be thought of as the *"perspective"* or
              *"client"* of a dependency *"request"*, due to be injected ASAP.

            :param key: Symbol that the caller would like to be provided by the
              DI pool, after exhausting more scope-specific options in the MRO
              chain.

            :raises AttributeError: If an attribute could not be found in the
              dependency pool, to simulate "not finding" a property on the
              origin ``klass``.

            :returns: Any result that the dependency pool can provide, from all
              known dependency resources available to ``klass``, available at
              ``key``. """

        # @TODO(sgammon): make things not jank here (don't always `collapse`)
        try:
          bridge = Proxy.Component.collapse(klass)
          if isinstance(bridge[key], tuple):  # pragma: no cover
            return getattr(*bridge[key])  # bridge key is (responder, attribute)
          return bridge[key]  # return value directly if it's not a tuple
        except KeyError:  # pragma: nocover
          raise AttributeError('Could not resolve attribute \'%s\''
                               ' on item \'%s\'.' % (key, klass))

      # inject properties onto MRO delegate, then construct
      return type.__new__(mcs.__class__, 'Delegate', (object,), {
        '__bridge__': None,
        '__getattr__': classmethod(injection_responder),
        '__metaclass__': mcs,
        '__repr__': mcs.__repr__,
        '__target__': target})

    def __repr__(cls):  # pragma: nocover

      """ Generate a string representation of the local ``Delegate``, including
          whatever class context we're running in, if any.

          :returns: String representation of a bound or root delegate. """

      return "<delegate root>" if not (
        cls.__target__) else "<delegate '%s'>" % cls.__target__.__name__

  @classmethod
  def bind(cls, target):

    """ Factory a new ``Delegate`` subclass, bound to the ``target`` context
        class.

        :param target: Subject class to bind and spawn a ``Delegate`` subtype
          for.

        :returns: Bound ``Delegate`` subtype, suitable for MRO mixing. """

    # wrap in Delegate class context as well
    return cls.__metaclass__.__new__(cls, target)


class Compound(type):

  """ Concrete class used as a metafactory for class structures that should
      enable attribute accesses for response from the DI pool.

      Use this class as a metaclass on items that should transparently work with
      attribute injection. """

  __seen__ = set()
  __delegate__ = None

  def mro(cls):

    """ Prepares MRO (Method Resolution Order) with a customized ``Delegate``
        that provides a view into the DI pool.

        :returns: Prepared MRO for a dependency-injected metatype, known as a
          ``Compound`` because it is essentially the product of itself and a
          collapsed set of ``Component`` injectables. """

    for base in cls.__bases__:
      # check if we've seen any of these bases
      if base not in (object, type) and base in cls.__class__.__seen__:
        break
    else:
      # never seen this before - roll in our delegate
      origin = type.mro(cls)
      delegate = Delegate.bind(cls)
      cls.__class__.__seen__.add(cls)

      return (
        [origin[0]] +
        [i for i in filter(lambda x: x not in (object, type), origin[1:])] +
        [y for y in filter(lambda z: z in (object, type), origin[1:])] +
        [delegate])

    return type.mro(cls)


class Bridge(object):

  """ Tiny utility class that can be used as a static ``Bridge`` into the DI
      pool. Suitable for use as an independent object, mounted wherever is
      convenient. """

  __metaclass__ = Compound



# -*- coding: utf-8 -*-

"""

  CLI utils
  ~~~~~~~~~

  toolset for making command-line based tools. useful in general,
  particularly useful in making app management utilities.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""


# stdlib
import sys
import argparse
import textwrap


## Globals
_root_tool = None


## == Embedded Metaclass == ##
class Tool(object):

  """ Meta parent class that applies an embedded metaclass to enforce/ transform
      encapsulated objects into :py:mod:`argparse` directives. """

  safe = False  # parse only known arguments
  parser = None  # local parser for this tool
  autorun = False  # should we auto-run parse?
  commands = None  # subcommands to this tool
  arguments = None  # arguments for this tool

  class __metaclass__(type):

    """ Bound utility metaclass that re-writes embedded comand classes
        on-the-fly, into :py:mod:`argparse`-provided objects. """

    tree = {}  # command parser tree
    parsers = {}  # all encountered parsers

    def __new__(mcs, name, bases, properties):

      """ Check to see if we're initializing a new subcommand class, and if we
          are, construct the appropriate subparser.

          :param name: Target class name.
          :param bases: Target class bases.
          :param properties: Class dict properties.

          :raises RuntimeError: If invalid tool bindings are expressed in a
            meta-initialized class (for instance, an argument without a name).

          :returns: Initialized class, transformed into additional objects
            provided by :py:mod:`argparse`. """

      # initialize `Tool` regularly to apply this metaclass downwards
      if name == 'Tool': return super(mcs, mcs).__new__(*(
                                    mcs, name, bases, properties))

      _subtools, _arguments = [], []
      for key, value in properties.viewitems():

        # is it a list of arguments?
        if isinstance(value, (list, tuple)) and key is 'arguments':

          def _add_argument(_parser, _flag, _cfg):  # pragma: no cover
            if isinstance(_flag, tuple):
              return _parser.add_argument(*_flag, **_cfg)
            return _parser.add_argument(_flag, **_cfg)

          for bundle in value:
            if len(bundle) == 2:
              _name, _config = bundle
              _arguments.append((_add_argument, _name, _config))
            else:
              if isinstance(bundle[-1], dict):
                positional, _config = bundle[0:-1], bundle[-1]
                _arguments.append((_add_argument, positional, _config))

        # is it a subtool?
        elif isinstance(value, type) and issubclass(value, Tool):

          def _add_subparser(root, obj, subparsers):  # pragma: no cover
            ## bind helptext from __doc__
            sub = subparsers.add_parser((getattr(obj, 'name') if (
              hasattr(obj, 'name')) else obj.__name__).lower(), **{
                'conflict_handler': 'resolve',
                'help': textwrap.dedent(getattr(obj, '__doc__').strip()) if (
                hasattr(obj, '__doc__') and (
                  getattr(obj, '__doc__') is not None)) else None})

            sub.set_defaults(func=obj.execute)
            return sub

          _subtools.append((value, _add_subparser))

        elif not key.startswith('__'):

          if not isinstance(value, classmethod) and callable(value):
            properties[key] = staticmethod(value)
          else:
            # let it through if it's marked as a classmethod
            properties[key] = value

      # construct class
      klass = super(mcs, mcs).__new__(mcs, name, bases, properties)

      # add to registered parsers
      mcs.parsers[name] = {
        'name': (properties['name'] if 'name' in properties else name).lower(),
        'description': textwrap.dedent(properties['__doc__']) if (
          '__doc__' in properties) else None,
        'implementation': klass,
        'objects': {
          'subtools': _subtools,
          'arguments': _arguments}}

      return klass

  def __init__(self, parser=None, autorun=False, safe=False):

    """ This initializer method is called at the tip of the toolchain tree
        (composed of :py:class:`Tool` classes) to start the process of
        initializing and constructing each :py:mod:`argparse` object.

        Execution cascades from the tip to sub- :py:class:`Tool`s, and then to
        arguments.

        :param parser: Existing ``argparse.ArgumentParser`` to use for parsing
          config arguments, or ``None`` if one should be created on-the-fly.

        :param autorun: ``bool`` flag indicating that the CLI tool should start
          up and try to parse flags immediately.

        :param safe: ``bool`` flag indicating that we should only parse and
          consider flags *explicitly defined* by the CLI tool, rather than
          letting through unknown flags. """

    global _root_tool

    self.autorun, self.safe = autorun, safe

    # lookup local config
    config = self.__metaclass__.parsers[self.__class__.__name__]

    if not parser:
      # start top-level argument parser
      parser = argparse.ArgumentParser(prog=(self.name if (
        hasattr(self, 'name')) else self.__class__.__name__).lower(),
        description=textwrap.dedent(self.__doc__.strip()) if not (
          (sys.flags.optimize > 1)) else '')
      if not _root_tool: _root_tool = parser

    self.parser = parser  # assign local parser

    # local args
    for _callable, flag, _config in (
      config.get('objects', {}).get('arguments', [])):
      _callable(parser, flag, _config)  # initialize each argument

    # local subtools
    if config.get('objects', {}).get('subtools', []):
      commands = parser.add_subparsers(dest='subcommand', title='bundled tools')
      for impl, _callable in config.get('objects', {}).get('subtools', []):
        subparser = _callable(_root_tool, impl, commands)  # initialize subtools
        setattr(self, (impl.name if (
          hasattr(impl, 'name')) else impl.__name__).lower(), impl(subparser))

    if autorun and safe:  # pragma: no cover
      self(*_root_tool.parse_known_args())
    elif autorun:  # pragma: no cover
      self(_root_tool.parse_args())

  @classmethod
  def execute(cls, arguments):  # pragma: no cover

    """ Execute the local ``Tool`` subclass against command-line-style args
        at ``arguments``, which are passed in from ``argparse``.

        :param arguments: Arguments object passed-in from ``argparse``, as an
          instance of :py:class:`argparse.Namespace`, resulting from a call to
          ``parser.parse_args``.

        :return: Must return a truthy value (such as ``True``) to indicate that
          whatever operation was requested completed successfully, or a falsy
          value otherwise (such as ``False``). Failure values produce a Unix
          exit code of ``1`` when running on the CLI, versus a code ``0`` for
          success/truthy values. """

    raise NotImplementedError('Command line tool "%s"'
                              ' is not implemented.' % repr(cls))

  def __call__(self, arguments, unknown=None):  # pragma: no cover

    """ Begins dispatching execution from a set of parsed arguments, as the
        product of a :py:meth:`parser.parse_args()` call.

        :param arguments: :py:class:`argparse.Namespace` object, resulting from
          ``parser.parse_args()``.

        :param unknown:

        :returns: Unix return code, suitable for passing directly to
          ``sys.exit()``. """

    try:
      # is it a subtool?
      if hasattr(arguments, 'func'):
        ## dispatch and return
        if self.safe:
          return_value = arguments.func(arguments, unknown)
        else:
          return_value = arguments.func(arguments)
      else:
        # no? ok
        if self.safe:
          return_value = self.execute(arguments, unknown)
        else:
          return_value = arguments.func(arguments)

    except Exception:
      raise
    return 1 if not return_value else 0



# -*- coding: utf-8 -*-

"""

  JSON protocol tests
  ~~~~~~~~~~~~~~~~~~~

  tests canteen's builtin JSON RPC protocol.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
      A copy of this license is included as ``LICENSE.md`` in
      the root of the project.

"""

# stdlib
import json

# canteen core
from canteen.core import Library

# canteen testing
from canteen.test import FrameworkTest


with Library('protorpc', strict=True) as (library, protorpc):

  # load messages library
  messages = library.load('messages')

  # JSON protocol
  from canteen.rpc.protocol import json as jsonrpc


  class SampleMessage(messages.Message):

    """ Sample ProtoRPC message. """

    string = messages.StringField(1)
    integer = messages.IntegerField(2)


  ## JSONProtocolTests
  # Tests the JSON-based RPC protocol.
  class JSONProtocolTests(FrameworkTest):

    """ Tests `rpc.protocol.json.JSON` """

    def test_json_construct(self):

      """ Test basic construction of JSON RPC protocol """

      jsonrpc.JSON()  # yep just that

    def test_json_encode_message(self):

      """ Test encoding RPC messages in JSON """

      # prepare message
      msg = SampleMessage(string='hi', integer=5)
      protocol = jsonrpc.JSON()

      # encode message
      result = protocol.encode_message(msg)

      # decode and test
      inflated = json.loads(result)

      # interrogate
      assert 'string' in inflated
      assert 'integer' in inflated
      assert inflated['string'] == 'hi'
      assert inflated['integer'] == 5

      return result, protocol

    def test_json_decode_message(self):

      """ Test decoding RPC messages from JSON """

      # prepare message
      encoded, protocol = self.test_json_encode_message()

      # decode serialized body
      decoded = protocol.decode_message(SampleMessage, encoded)

      # interrogate
      assert decoded.string == 'hi'
      assert decoded.integer == 5



# -*- coding: utf-8 -*-

"""

  inmemory adapter tests
  ~~~~~~~~~~~~~~~~~~~~~~

  tests canteen's builtin inmemory DB engine.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
      A copy of this license is included as ``LICENSE.md`` in
      the root of the project.

"""

# stdlib
import datetime

# canteen model API
from canteen import model
from canteen.model.adapter import inmemory

# abstract test bases
from .test_abstract import DirectedGraphAdapterTests


## Globals
_target = lambda k: k.flatten(True)[1]


class InMemoryModel(model.Model):

  """ Test model. """

  __adapter__ = inmemory.InMemoryAdapter

  string = basestring, {'required': True}
  integer = int, {'repeated': True}
  number = int, {'repeated': False}
  floating = float, {'required': False}
  date = datetime.datetime


class InMemoryAdapterTests(DirectedGraphAdapterTests):

  """ Tests `model.adapter.inmemory` """

  __abstract__ = False
  subject = inmemory.InMemoryAdapter

  def test_construct(self):

    """ Test basic construction of `InMemoryAdapter` """

    x = inmemory.InMemoryAdapter()
    assert x
    assert isinstance(x, inmemory.InMemoryAdapter)

  def test_invalid_get(self):

    """ Test requesting a key that doesn't exist """

    # get missing entity
    model_key = model.Key("Sample", "_____")
    entity = model_key.get()
    assert entity is None

  def test_named_entity_get_put(self):

    """ Test putting and getting an entity with a named key """

    # put entity
    m = InMemoryModel(
      key=model.Key(InMemoryModel.kind(), "NamedEntity"),
      string="suphomies",
      integer=[4, 5, 6, 7])
    m_k = m.put()

    assert _target(m_k) in inmemory._metadata['__key__']

    entities = []

    # simulate getting entity at key level
    explicit_key = model.Key(InMemoryModel.kind(), "NamedEntity")
    entity = explicit_key.get()
    entities.append(entity)

    # simulate getting entity at model level
    explicit_entity = InMemoryModel.get(name="NamedEntity")
    entities.append(explicit_entity)

    # test urlsafe-d key model-level get()
    urlsafed_entity = InMemoryModel.get(key=explicit_key.urlsafe())
    entities.append(urlsafed_entity)

    # test raw-d key model-level get()
    flattened = explicit_key.flatten(False)
    rawd_entity = InMemoryModel.get(key=flattened[1:])
    entities.append(rawd_entity)

    for entity in entities:
      # make sure things match on key level
      assert entity.string == "suphomies"
      assert len(entity.integer) == 4
      assert entity.key.id == "NamedEntity"
      assert entity.key.kind == InMemoryModel.kind()

  def test_id_entity_get_put(self):

    """ Test putting and getting an entity with an ID'd key """

    # put entity
    m = InMemoryModel(string="hello", integer=[1, 2, 3])
    m_k = m.put()

    assert _target(m_k) in inmemory._metadata['__key__']

    # make sure flags match
    assert m_k == m.key
    assert m_k.__persisted__
    assert m.__persisted__
    assert (not m.__dirty__)

    # simulate getting entity via urlsafe
    entity = InMemoryModel.get(m_k)

    # make sure things match
    assert entity.string == "hello"
    assert len(entity.integer) == 3
    assert entity.key.kind == InMemoryModel.kind()

  def test_delete_existing_entity_via_key(self):

    """ Test deleting an existing entity via `Key.delete()` """

    # put entity
    m = InMemoryModel(string="hello", integer=[1, 2, 3])
    m_k = m.put()

    # make sure it's known
    assert _target(m_k) in inmemory._metadata['__key__']

    # delete it
    res = m_k.delete()

    # make sure it's unknown and gone
    assert res
    assert _target(m_k) not in inmemory._metadata['__key__']

  def test_delete_existing_entity_via_model(self):

    """ Test deleting an existing entity via `Model.delete()` """

    # put entity
    m = InMemoryModel(string="hello", integer=[1, 2, 3])
    m_k = m.put()

    # make sure it's known
    assert _target(m_k) in inmemory._metadata['__key__']

    # delete it
    res = m.delete()

    # make sure it's unknown and gone
    assert res
    assert _target(m_k) not in inmemory._metadata['__key__']

  def test_delete_invalid_entity(self):

    """ Test deleting an invalid entity """

    # manufacture a key that shouldn't exist...
    m_k = model.Key("SampleKind", "____InvalidKey____")

    # make sure it's unknown
    assert _target(m_k) not in inmemory._metadata['__key__']

    # delete it
    res = m_k.delete()

    # make sure it's unknown, and we couldn't delete it
    assert (not res)
    assert _target(m_k) not in inmemory._metadata['__key__']

  def test_allocate_ids(self):

    """ Test allocating one and numerous ID's """

    # try allocating one ID
    next = inmemory.InMemoryAdapter.allocate_ids("Sample", 1)
    assert isinstance(next, int)

    # try allocating 10 ID's
    next_range = [i for i in inmemory.InMemoryAdapter.allocate_ids(*(
      model.Key, "Sample", 10))()]
    assert len(next_range) == 10
    for i in next_range:
      assert isinstance(i, int)

  def test_equality_query(self):

    """ Test equality queries with `InMemoryAdapter` """

    # make some models
    m = [
      InMemoryModel(string="soop", integer=[1, 2, 3]),
      InMemoryModel(string="soop", integer=[1, 2, 3]),
      InMemoryModel(string="soop", integer=[1, 2, 3])]

    for _m in m: _m.put()

    # single get
    q = InMemoryModel.query().filter(InMemoryModel.string == "soop")
    result = q.get()
    assert result.string == "soop"

    # submit query
    q = InMemoryModel.query().filter(InMemoryModel.string == "soop")
    result = q.fetch(limit=50)

    for r in result:
      assert r.string == "soop"

  def test_inequality_query(self):

    """ Test inequality queries with `InMemoryAdapter` """

    # make some models
    m = [
      InMemoryModel(string="soop", integer=[1, 2, 3]),
      InMemoryModel(string="soop", integer=[1, 2, 3]),
      InMemoryModel(string="soop", integer=[1, 2, 3]),
      InMemoryModel(string="sploop", integer=[1, 2])]

    for _m in m: _m.put()

    # submit query
    q = InMemoryModel.query().filter(InMemoryModel.string != "sploop")
    result = q.fetch(limit=50)

    assert len(result) > 0, (
      "got no results for inequality query"
      " (got '%s' from adapter '%s')" % (result, self.subject))

    for r in result:
      assert r.string != "sploop"

  def test_range_query(self):

    """ Test range queries with `InMemoryAdapter` """

    now = datetime.datetime.now()

    # make some models
    m = [
      InMemoryModel(string="soop", date=now + datetime.timedelta(days=1)),
      InMemoryModel(string="soop", date=now + datetime.timedelta(days=2)),
      InMemoryModel(string="soop", date=now + datetime.timedelta(days=3))]

    for _m in m: _m.put()

    # submit query
    q = InMemoryModel.query().filter(InMemoryModel.date > now)
    result = q.fetch(limit=50)

    for result_i in result:
      assert result_i.date > now

  def test_ancestry_query(self):

    """ Test ancestry queries with `InMemoryAdapter` """

    root = model.Key(InMemoryModel, 'heyo')
    _key = lambda x: model.Key(InMemoryModel, x, parent=root)
    child1 = _key('child')
    child2 = _key('child2')

    # make some models
    m = [
      InMemoryModel(key=root, string='soop'),
      InMemoryModel(key=child1, string='soop'),
      InMemoryModel(key=child2, string='soop')]

    for _m in m: _m.put()

    # submit query
    q = InMemoryModel.query(ancestor=root, limit=50)
    result = q.fetch()

    assert len(result) == 2

  def test_compound_query(self):

    """ Test compound queries with `InMemoryAdapter` """

    now = datetime.datetime.now()
    root = model.Key(InMemoryModel, 'hi')
    _key = lambda x: model.Key(InMemoryModel, x, parent=root)
    child1 = _key('child1')
    child2 = _key('child2')
    child3 = _key('child3')

    # make some models
    m = [
      InMemoryModel(key=root,
                    string="hithere",
                    date=now + datetime.timedelta(days=1)),
      InMemoryModel(key=child1,
                    string="hithere",
                    date=now + datetime.timedelta(days=2)),
      InMemoryModel(key=child2,
                    string="hithere",
                    date=now + datetime.timedelta(days=3)),
      InMemoryModel(key=child3,
                    string="noway",
                    date=now + datetime.timedelta(days=4))]

    for _m in m: _m.put()

    # submit query
    q = InMemoryModel.query(limit=50)
    q.filter(InMemoryModel.string == "hithere")
    q.filter(InMemoryModel.date > now)
    result = q.fetch()

    assert len(result) == 3

    # submit query
    q = InMemoryModel.query(ancestor=root, limit=50)
    q.filter(InMemoryModel.date > now)
    result = q.fetch()

    assert len(result) == 3

    # submit query
    q = InMemoryModel.query(ancestor=root, limit=50)
    q.filter(InMemoryModel.date > now)
    q.filter(InMemoryModel.string == "hithere")
    result = q.fetch()

    assert len(result) == 2

  def test_ascending_sort_string(self):

    """ Test an ASC sort on a string property with `InMemoryAdapter` """

    root = model.Key(InMemoryModel, 'sorted-string')
    _key = lambda x: model.Key(InMemoryModel, x, parent=root)
    child1 = _key('child1')
    child2 = _key('child2')
    child3 = _key('child3')
    child4 = _key('child4')

    # make some models
    m = [
      InMemoryModel(key=child1, string="aardvark"),
      InMemoryModel(key=child2, string="blasphemy"),
      InMemoryModel(key=child3, string="xylophone"),
      InMemoryModel(key=child4, string="yompin")
    ]

    for _m in m: _m.put()

    # submit query
    q = InMemoryModel.query(ancestor=root, limit=50).sort(+InMemoryModel.string)
    result = q.fetch()

    assert len(result) == 4

    for l, r in zip(result, reversed(("aardvark",
                                      "blasphemy",
                                      "xylophone",
                                      "yompin"))):
      assert l.string == r

  def test_descending_sort_string(self):

    """ Test a DSC sort on a string property with `InMemoryAdapter` """

    root = model.Key(InMemoryModel, 'sorted-string-2')
    _key = lambda x: model.Key(InMemoryModel, x, parent=root)
    child1 = _key('child1')
    child2 = _key('child2')
    child3 = _key('child3')
    child4 = _key('child4')

    # make some models
    m = [
      InMemoryModel(key=child1, string="aardvark"),
      InMemoryModel(key=child2, string="blasphemy"),
      InMemoryModel(key=child3, string="xylophone"),
      InMemoryModel(key=child4, string="yompin")]

    for _m in m: _m.put()

    # submit query
    q = InMemoryModel.query(ancestor=root, limit=50).sort(-InMemoryModel.string)
    result = q.fetch()

    assert len(result) == 4

    for l, r in zip(result, ("aardvark", "blasphemy", "xylophone", "yompin")):
      assert l.string == r

  def test_ascending_sort_integer(self):

    """ Test an ASC sort on an integer property with `InMemoryAdapter` """

    root = model.Key(InMemoryModel, 'sorted-int')
    _key = lambda x: model.Key(InMemoryModel, x, parent=root)
    child1 = _key('child1')
    child2 = _key('child2')
    child3 = _key('child3')
    child4 = _key('child4')

    # make some models
    m = [
      InMemoryModel(key=child1, string="aardvark", number=5),
      InMemoryModel(key=child2, string="blasphemy", number=6),
      InMemoryModel(key=child3, string="xylophone", number=7),
      InMemoryModel(key=child4, string="yompin", number=8)]

    for _m in m: _m.put()

    # submit query
    q = InMemoryModel.query(ancestor=root, limit=50).sort(+InMemoryModel.number)
    result = q.fetch()

    assert len(result) == 4

    for l, r in zip(result, (5, 6, 7, 8)):
      assert l.number == r

  def test_descending_sort_integer(self):

    """ Test a DSC sort on an integer property with `InMemoryAdapter` """

    root = model.Key(InMemoryModel, 'sorted-int-2')
    _key = lambda x: model.Key(InMemoryModel, x, parent=root)
    child1 = _key('child1')
    child2 = _key('child2')
    child3 = _key('child3')
    child4 = _key('child4')

    # make some models
    m = [
      InMemoryModel(key=child1, string="aardvark", number=5),
      InMemoryModel(key=child2, string="blasphemy", number=6),
      InMemoryModel(key=child3, string="xylophone", number=7),
      InMemoryModel(key=child4, string="yompin", number=8)]

    for _m in m: _m.put()

    # submit query
    q = InMemoryModel.query(ancestor=root, limit=50).sort(-InMemoryModel.number)
    result = q.fetch()

    assert len(result) == 4

    for l, r in zip(result, reversed((5, 6, 7, 8))):
      assert l.number == r

  def test_ascending_sort_float(self):

    """ Test an ASC sort on a float property with `InMemoryAdapter` """

    root = model.Key(InMemoryModel, 'sorted-float')
    _key = lambda x: model.Key(InMemoryModel, x, parent=root)
    child1 = _key('child1')
    child2 = _key('child2')
    child3 = _key('child3')
    child4 = _key('child4')

    # make some models
    m = [
      InMemoryModel(key=child1, string="aardvark", floating=5.5),
      InMemoryModel(key=child2, string="blasphemy", floating=6.5),
      InMemoryModel(key=child3, string="xylophone", floating=7.5),
      InMemoryModel(key=child4, string="yompin", floating=8.5)]

    for _m in m: _m.put()

    # submit query
    q = (
     InMemoryModel.query(ancestor=root, limit=50).sort(+InMemoryModel.floating))
    result = q.fetch()

    assert len(result) == 4

    for l, r in zip(result, (5.5, 6.5, 7.5, 8.5)):
      assert l.floating == r

  def test_descending_sort_float(self):

    """ Test a DSC sort on a float property with `InMemoryAdapter` """

    root = model.Key(InMemoryModel, 'sorted-float')
    _key = lambda x: model.Key(InMemoryModel, x, parent=root)
    child1 = _key('child1')
    child2 = _key('child2')
    child3 = _key('child3')
    child4 = _key('child4')

    # make some models
    m = [
      InMemoryModel(key=child1, string="aardvark", floating=5.5),
      InMemoryModel(key=child2, string="blasphemy", floating=6.5),
      InMemoryModel(key=child3, string="xylophone", floating=7.5),
      InMemoryModel(key=child4, string="yompin", floating=8.5)]

    for _m in m: _m.put()

    # submit query
    q = (
     InMemoryModel.query(ancestor=root, limit=50).sort(-InMemoryModel.floating))
    result = q.fetch()

    assert len(result) == 4

    for l, r in zip(result, reversed((5.5, 6.5, 7.5, 8.5))):
      assert l.floating == r

  def test_ascending_sort_datetime(self):

    """ Test an ASC sort on a `datetime` property with `InMemoryAdapter` """

    now = datetime.datetime.now()
    later = lambda n: now + datetime.timedelta(days=n)
    root = model.Key(InMemoryModel, 'sorted-datetime')
    _key = lambda x: model.Key(InMemoryModel, x, parent=root)
    child1 = _key('child1')
    child2 = _key('child2')
    child3 = _key('child3')
    child4 = _key('child4')

    # make some models
    m = [
      InMemoryModel(key=child1, string="aardvark", date=later(1)),
      InMemoryModel(key=child2, string="blasphemy", date=later(2)),
      InMemoryModel(key=child3, string="xylophone", date=later(3)),
      InMemoryModel(key=child4, string="yompin", date=later(4))]

    for _m in m: _m.put()

    # submit query
    q = InMemoryModel.query(ancestor=root, limit=50).sort(+InMemoryModel.date)
    result = q.fetch()

    assert len(result) == 4

    for l, r in zip(result, (later(1), later(2), later(3), later(4))):
      assert l.date == r

  def test_descending_sort_datetime(self):

    """ Test a DSC sort on a `datetime` property with `InMemoryAdapter` """

    now = datetime.datetime.now()
    later = lambda n: now + datetime.timedelta(days=n)
    root = model.Key(InMemoryModel, 'sorted-datetime')
    _key = lambda x: model.Key(InMemoryModel, x, parent=root)
    child1 = _key('child1')
    child2 = _key('child2')
    child3 = _key('child3')
    child4 = _key('child4')

    # make some models
    m = [
      InMemoryModel(key=child1, string="aardvark", date=later(1)),
      InMemoryModel(key=child2, string="blasphemy", date=later(2)),
      InMemoryModel(key=child3, string="xylophone", date=later(3)),
      InMemoryModel(key=child4, string="yompin", date=later(4))]

    for _m in m: _m.put()

    # submit query
    q = InMemoryModel.query(ancestor=root, limit=50).sort(-InMemoryModel.date)
    result = q.fetch()

    assert len(result) == 4

    for l, r in zip(result, reversed((later(1), later(2), later(3), later(4)))):
      assert l.date == r



# -*- coding: utf-8 -*-

"""

  model queries
  ~~~~~~~~~~~~~

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
      A copy of this license is included as ``LICENSE.md`` in
      the root of the project.

"""


__version__ = 'v5'


# stdlib
import abc
import operator

# datastructures
from canteen.util.struct import (EMPTY,
                                 Sentinel)


## Globals / Constants

# Filter components
_TARGET_KEY = Sentinel('KEY')
PROPERTY = Sentinel('PROPERTY')
KEY_KIND = Sentinel('KEY_KIND')
KEY_ANCESTOR = Sentinel('KEY_ANCESTOR')
EDGES = Sentinel('EDGES')
NEIGHBORS = Sentinel('NEIGHBORS')

# Sort directions
ASCENDING = ASC = Sentinel('ASCENDING')
DESCENDING = DSC = Sentinel('DESCENDING')

# Query operators
EQUALS = EQ = Sentinel('EQUALS')
NOT_EQUALS = NE = Sentinel('NOT_EQUALS')
LESS_THAN = LT = Sentinel('LESS_THAN')
LESS_THAN_EQUAL_TO = LE = Sentinel('LESS_THAN_EQUAL_TO')
GREATER_THAN = GT = Sentinel('GREATER_THAN')
GREATER_THAN_EQUAL_TO = GE = Sentinel('GREATER_THAN_EQUAL_TO')
CONTAINS = IN = Sentinel('CONTAINS')

# Logic operators
AND = Sentinel('AND')
OR = Sentinel('OR')

# Operator Constants
_operator_map = {
  EQUALS: operator.eq,
  NOT_EQUALS: operator.ne,
  LESS_THAN: operator.lt,
  LESS_THAN_EQUAL_TO: operator.le,
  GREATER_THAN: operator.gt,
  GREATER_THAN_EQUAL_TO: operator.ge,
  CONTAINS: operator.contains,
  AND: operator.__and__,
  OR: operator.__or__}

_operator_strings = {
  EQUALS: '==',
  NOT_EQUALS: '!=',
  LESS_THAN: '<',
  LESS_THAN_EQUAL_TO: '<=',
  GREATER_THAN: '>',
  GREATER_THAN_EQUAL_TO: '>=',
  CONTAINS: 'CONTAINS',
  AND: 'AND',
  OR: 'OR'}


class QueryOptions(object):

  """ Holds a re-usable set of options for a :py:class:`Query`. """

  # == Options == #
  options = frozenset((
    '_keys_only',
    '_ancestor',
    '_limit',
    '_offset',
    '_projection',
    '_hint',
    '_plan',
    '_cursor'))

  __slots__ = frozenset(('__explicit__',)) | options
  option_names = frozenset(('_'.join(opt.split('_')[1:]) for opt in options))

  # == Option Defaults == #
  _defaults = {
    '_keys_only': False,
    '_ancestor': None,
    '_limit': -1,
    '_offset': 0,
    '_projection': None,
    '_hint': None,
    '_plan': None,
    '_cursor': None}

  ## == Internal Methods == ##
  def __init__(self, **kwargs):

    """ Initialize this :py:class:`QueryOptions`. Map ``kwargs`` into local data
        properties that are abstracted behind getters/setters at the class
        level.

        :param **kwargs: Keyword argument options. Valid keys are listed in
          :py:attr:`__slots__`.

        :raises AttributeError: In the case that an invalid config key is found
          in ``kwargs``. Passed-up from :py:meth:`_set_option`.

        :returns: Nothing, as this is a constructor. """

    self.__explicit__ = False  # initialize explicit flag
    map(lambda bundle: self._set_option(*bundle),
        map(lambda slot: (slot, kwargs.get(slot, EMPTY)), self.option_names))

  def __iter__(self):

    """ Iterate over options specified in this :py:class:`QueryOptions` object.
        Imitates the ``dict.iteritems`` interface, in that it yields
        ``(k, v)`` pairs. """

    for option in self.option_names:
      yield option, self._get_option(option)

  def __enter__(self):

    """ Enter an iteration mode whereby default *values* are omitted in favor of
        a sentinel that indicates empty fields.

        :returns: ``self``, in case an ``as`` block is used. """

    self.__explicit__ = True
    return self

  def __exit__(self, type, value, traceback):

    """ Exit the special iteration mode whereby default values are omitted.

        :returns: ``True`` in the case of a successful (read, non-excepting)
          enclosed block, else ``False`` to propagate the exception. """

    self.__explicit__ = False
    if traceback: return False
    return True

  def __repr__(self):

    """ Generate a pleasant string representation of this
        :py:class:`QueryOptions` object.

        :returns: Pleasant ``str`` label for this object. """

    properties = []
    with self:
      for k, v in self:
        if v is not EMPTY:
          properties.append((k, v))

    return "QueryOptions(%s)" % (
      ", ".join(map(lambda (k, v): "=".join((k, str(v))), properties)))

  ## == Protected Methods == ##
  def _set_option(self, name, value=EMPTY, _setter=False):

    """ Set the value of an option local to this :py:class:`QueryOptions`
        object. Calling without a ``value`` (which defaults to ``None``) resets
        the target key's value.

        :param name: Name (``str``) of the internal property we're setting.
        :param value: Value to set the property to. Defaults to ``None``.
        :param _setter: Internal ``bool``-type param indicating we're accessing
          this from a setter, so we should not return a value.

        :raises ValueError: If ``name`` is not a ``basestring`` descendent.

        :raises AttributeError: If ``name`` is not a valid internal property
          name.

        :returns: ``self``, for chainability. """

    if not isinstance(name, basestring):
      raise ValueError('Argument `name` of `_set_option` must'
                       ' be a string internal propery name.'
                       ' Got: "%s".' % str(name))  # pragma: no cover

    # build internal name
    name = '_' + unicode(name) if name[0] != '_' else unicode(name)

    if name not in self.__slots__:
      raise AttributeError('`QueryOptions` object has no option by'
                           ' the name "%s".' % name)

    setattr(self, name, value)  # set value and return
    if not _setter:
      return self

  def _get_option(self, name, default=EMPTY):

    """ Get the value of an option local to this :py:class:`QueryOptions`
        object.

        :param name: Name (``str``) of the internal property we're getting.

        :param default: Default value to return if no value was found at
          ``name``. Defaults to ``None``.

        :raises ValueError: If ``name`` is not a ``basestring`` descendent.

        :raises AttributeError: If ``name`` is not a valid internal property
          name.

        :returns: Configuration value at ``name``, or ``default`` if no value
          was found. """

    if not isinstance(name, basestring):  # pragma: no cover
      raise ValueError('Argument `name` of `_get_option` must'
                       ' be a string internal property name.'
                       ' Got: "%s".' % name)

    name = '_%s' % name  # build internal name

    if name not in self.__slots__:
      raise AttributeError('`QueryOptions` object has no option by'
                           ' the name "%s".' % name)  # pragma: no cover

    val = getattr(self, name)  # get value

    # return default value if empty slot was found, otherwise return value
    if val is EMPTY:
      if self.__explicit__:
        return EMPTY
      if default is EMPTY:
        return self._defaults.get(name, None)
      return default  # pragma: no cover
    return val

  def overlay(self, other, override=False):

    """ Combine this :py:class:`QueryOptions` object with another one, by
        merging the ``other`` object's settings into this one.

        :param other: Other :py:class:`QueryOptions` object to collect overriden
          property values from.

        :param override:

        :returns: Resulting merged object, which is simply ``self``. """

    with other:
      with self:
        for option, value in other:
          if value is not EMPTY:
            this_value = self._get_option(option)
            if (this_value is EMPTY) or override:
              self._set_option(option, value)
    return self


  ## == Public Properties == ##

  # ``keys_only`` flag - return keys instead of entities
  keys_only = property(lambda self: self._get_option('keys_only'))

  # ``ancestor`` filter - restrict results by key ancestry
  ancestor = property(lambda self: self._get_option('ancestor'))

  # ``limit`` - return a limited number of query results
  limit = property(lambda self: self._get_option('limit'))

  # ``offset`` - skip an amount of records before building results
  offset = property(lambda self: self._get_option('offset'))

  # ``projection`` - retrieve entity values from indexes while fulfilling query
  projection = property(lambda self: self._get_option('projection'))

  # ``plan`` - cached plan to fulfill the query (optional)
  plan = property(lambda self: self._get_option('plan'),
                  lambda self, v: self._set_option('plan', v, _setter=True))

  # ``cursor`` - result cursor, for paging or long queries (optional)
  cursor = property(lambda self: self._get_option('cursor'),
                    lambda self, v: self._set_option('cursor', v, _setter=True))


class AbstractQuery(object):

  """ Specifies base structure and interface for all query classes. """

  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def filter(self, expression):  # pragma: no cover

    """ Add a filter to the active :py:class:`Query` at ``self``.

        :param expression: Filter expression to apply during query planning.
        :raises NotImplementedError: Always, as this method is abstract.
        :returns: ``self``, for chainability. """

    raise NotImplementedError('`filter` is abstract and may not'
                              ' be invoked directly.')  # pragma: no cover

  @abc.abstractmethod
  def sort(self, expression):  # pragma: no cover

    """ Add a sort directive to the active :py:class:`Query` at ``self``.

        :param expression: Sort expression to apply to the target result set.
        :raises NotImplementedError: Always, as this method is abstract.
        :returns: ``self``, for chainability. """

    raise NotImplementedError('`sort` is abstract and may not'
                              ' be invoked directly.')  # pragma: no cover

  @abc.abstractmethod
  def hint(self, directive):  # pragma: no cover

    """ Pass a hint to the query-planning subsystem for how this query could
        most efficiently be satisfied.

        :param expression: Hint expression to take into consideration.
        :raises NotImplementedError: Always, as this method is abstract.
        :returns: ``self``, for chainability. """

    raise NotImplementedError('`hint` is abstract and may not'
                              ' be invoked directly.')  # pragma: no cover

  @abc.abstractmethod
  def fetch(self, **options):  # pragma: no cover

    """ Fetch results for a :py:class:`Query`, via the underlying driver's
        :py:meth:`execute_query` method.

        :param **options: Query options to build into a :py:class:`QueryOptions`
          object.

        :raises NotImplementedError: Always, as this method is abstract.

        :returns: Iterable (``list``) of matching :py:class:`model.Model`
          entities (or :py:class:`model.Key` objects if ``keys_only`` is
          truthy) yielded from current :py:class:`Query`, or an empty ``list``
          if no results were found. """

    raise NotImplementedError('`fetch` is abstract and may not'
                              ' be invoked directly.')  # pragma: no cover

  @abc.abstractmethod
  def get(self, **options):  # pragma: no cover

    """ Get a single result (by default, the first) matching a
        :py:class:`Query`.

        :param **options: Query options to build into a
          :py:class:`QueryOptions` object.

        :raises NotImplementedError: Always, as this method is abstract.

        :returns: Single result :py:class:`model.Model` (or
          :py:class:`model.Key` if ``keys_only`` is truthy) matching the current
          :py:class:`Query`, or ``None`` if no matching entities were found. """

    raise NotImplementedError('`fetch` is abstract and may not'
                              ' be invoked directly.')  # pragma: no cover


class GraphQueryOptions(QueryOptions):
  """ Specifies Graph-related query options in addition to standard ``Query``
      properties. """

  # == Options == #
  options = QueryOptions.options | frozenset((
    '_base',))

  __slots__ = QueryOptions.options | options

  option_names = frozenset(('_'.join(opt.split('_')[1:]) for opt in options))

  # == Option Defaults == #
  _defaults = dict(QueryOptions._defaults, **{
    '_graph_base': None})


  # # == Public Properties == ##

  # ``base`` - graph-specific base key
  base = property(lambda self: self._get_option('base'))


class Query(AbstractQuery):

  """ Top-level class representing a specification for a query across data
      accessible to the canteen model layer, using an adapter that supports the
      :py:class:`IndexedModelAdapter` interface. """

  kind = None  # model kind
  sorts = None  # sort directives
  options = None  # attached query options
  filters = None  # filter directives

  def __init__(self, kind=None, filters=None, sorts=None, **kwargs):

    """ Initialize this :py:class:`Query`, assigning any properties/config
        passed in via ``kwargs`` and such.

        :param kind: Model kind that we wish to query upon, or ``None`` to
          specify a *kindless* query (meaning that it matches *all* available
          model kinds, if implemented).

        :param filters: Array of ``Filter`` objects to add to the newly-prepared
          ``Query`` object. ``Filter``s can also be added via ``Query.filter``.

        :param sorts: Array of ``Sort`` objects to add to the newly-prepared
          ``Query`` object. ``Sort``s can also be added via ``Query.sort``.

        :param **kwargs: Additional options to pass to ``QueryOptions``. If an
          item exists at ``kwargs['options']``, it is used *in place* of
          generated query options. """

    filters = filters and ([filters] if not (
      isinstance(filters, (list, tuple))) else filters) or None

    sorts = sorts and ([sorts] if not (
      isinstance(sorts, (list, tuple))) else sorts) or None

    options = kwargs.get('options', QueryOptions(**kwargs))
    self.kind, self.filters, self.sorts, self.options = (
      kind, filters or [], sorts or [], options)

  def __repr__(self):

    """ Generate a string representation of this :py:class:`Query`.

        :returns: String representation of the current :py:class:`Query`. """

    return "Query(%s, filter=%s, sort=%s, options=%s)" % (
      self.kind.kind(),
      '[' + ','.join((str(f) for f in self.filters)) + ']',
      '[' + ','.join((str(s) for s in self.sorts)) + ']',
      self.options.__repr__()
    )

  # @TODO(sgammon): async methods to execute
  def _execute(self, options=None, adapter=None, **kwargs):

    """ Internal method to execute a query, optionally along with some override
        options.

        .. note: This method will eventually accompany an async equivalent,
             which this will make use of runtime async tools under-the-hood.

        :param options:
        :param adapter:

        :param **kwargs: Keyword arguments of query config (i.e. valid and
          registered on :py:class:`QueryOptions`) to pass to the options object
          built to execute the query.

        :raises AttributeError: In the case that an invalid/unknown query
          configuration key is encountered. Passed up from
          :py:class:`QueryOptions`.

        :raises NotImplementedError: In the case that a ``kindless`` or
          ``projection`` query is encountered, as those features are not yet
          supported.

        :returns: Synchronously-retrieved results to this :py:class:`Query`. """

    ## build query options
    if self.options and options:
      options = self.options.overlay(options)
    else:
      options = (options or (
        kwargs.get('options', (
          QueryOptions(**kwargs) if kwargs else self.options))))

    ## fail for projection queries
    if options.projection:
      raise NotImplementedError('Projection queries are not'
                                ' yet supported.')  # pragma: no cover

    if adapter:
      return adapter._execute_query(self)

    if self.kind:  # kinded query

      # delegate to driver
      return self.kind.__adapter__._execute_query(self)

    else:

      # kindless queries are not yet supported
      raise NotImplementedError('Kindless queries are not'
                                ' yet supported.')  # pragma: no cover

  def filter(self, expression):

    """ Add a filter to this :py:class:`Query`.

        :param expression: Expression descriptor of type :py:class:`Filter`.

        :raises:
        :returns: """

    if isinstance(expression, Filter):
      self.filters.append(expression)
    else:  # pragma: no cover
      raise NotImplementedError('Query method `filter` does not yet support'
                                ' non-`Filter` component types.')
    return self

  def sort(self, expression):

    """ Add a sort order to this :py:class:`Query`.

        :param expression: Expression descriptor of type :py:class:`Sort`.

        :raises:
        :returns: """

    from canteen import model

    if isinstance(expression, model.Property):  # pragma: no cover
      # default to descending sort
      self.sorts.append(-expression)
    if isinstance(expression, Sort):
      self.sorts.append(expression)
    else:  # pragma: no cover
      raise NotImplementedError('Query method `sort` does not yet support non-'
                                '`Sort` component types.')
    return self

  def hint(self, directive):  # pragma: no cover

    """ Provide an external hint to the query planning logic about how to plan
        the query.

        Currently stubbed.

        :param directive:

        :raises NotImplementedError: Always,
        as this method is currently stubbed.

        :returns: ``self``, for chainability. """

    # @TODO(sgammon): fill out query hinting logic
    raise NotImplementedError('Query method `hint` is currently stubbed.')

  def get(self, adapter=None, **options):

    """ Get a single result (by default, the first) matching a
        :py:class:`Query`.

        :param **options: Accepts any valid and registered options on
          :py:class:`QueryOptions`.

        :param adapter: Adapter to use for the ``get`` operation.

        :returns: Single result :py:class:`model.Model` (or
          :py:class:`model.Key` if ``keys_only`` is truthy) matching the current
          :py:class:`Query`, or ``None`` if no matching entities were found. """

    options['limit'] = 1  # always has a limit of 1
    result = self._execute(QueryOptions(**options), adapter)
    return result[0] if result else result

  def fetch(self, adapter=None, **options):

    """ Fetch results for the currently-built :py:class:`Query`, executing it
        across the attached ``kind``'s attached model adapter.

        :param **options: Accepts any valid and registered options on
          :py:class:`QueryOptions`.

        :returns: Iterable (``list``) of matching model entities. """

    return self._execute(
      options=QueryOptions(**options) if options else None,
      adapter=adapter)

  def fetch_page(self, **options):

    """ Fetch a page of results, potentially as the next in a sequence of page
        requests.

        :param **options: Accepts any valid and registered options on
          :py:class:`QueryOptions`.

        :raises:

        :returns: """

    # @TODO(sgammon): build out paging support
    raise NotImplementedError('Query method `fetch_page`'
                              ' is currently stubbed.')  # pragma: no cover


class QueryComponent(object):

  """ Top-level abstract class for a component of a :py:class:`Query`, which is
      usually an attached specification like a :py:class:`Filter` or
      :py:class:`Sort`. """

  __metaclass__ = abc.ABCMeta

  ## == Component State == ##
  kind = None  # kind of property
  target = None  # property to operate on
  operator = None  # operator selection

  ## == Constants == ##
  PROPERTY = PROPERTY
  KEY_KIND = KEY_KIND
  KEY_ANCESTOR = KEY_ANCESTOR


class Filter(QueryComponent):

  """ Query component specification parent for a generic filter, used to
      traverse indexes and find entities to return that match. """

  generic_name = 'Property'

  ## == Filter State == ##
  value = None  # value to match
  chain = None  # chained AND and OR filters
  sub_operator = None  # subquery operator, if any (makes this a subfilter)

  ## == Operators == ##
  EQUALS = EQ = EQUALS
  NOT_EQUALS = NEQ = NOT_EQUALS
  LESS_THAN = LT = LESS_THAN
  LESS_THAN_EQUAL_TO = LE = LESS_THAN_EQUAL_TO
  GREATER_THAN = GT = GREATER_THAN
  GREATER_THAN_EQUAL_TO = GE = GREATER_THAN_EQUAL_TO
  CONTAINS = IN = CONTAINS

  def __init__(self, prop, value,
                            AND=None,
                            OR=None,
                            type=PROPERTY,
                            operator=EQUALS):

    """ Initialize this :py:class:`Filter`.

        :param prop: Property object to filter against.

        :param value: Value to filter entity ``prop`` containers against.

        :param AND: Subfilters to chain with a logical ``AND``.

        :param OR: Subfilters to chain with a logical ``OR``.

        :param type: Type of filter to create, defaults to ``PROPERTY``,
          meaning it is a filter against a ``value`` in properties contained by
          stored entities.

        :param operator: Operator to use for comparing ``prop`` values against
          ``value``. Can be one of the following options:
          - ``EQUALS``, equating to ``x == y``
          - ``NOT_EQUALS``, equating to ``x != y``
          - ``LESS_THAN``, equating to ``x < y``
          - ``LESS_THAN_EQUAL_TO``, equating to ``x <= y``
          - ``GREATER_THAN``, equating to ``x > y``
          - ``GREATER_THAN_EQUAL_TO``, equating to ``x >= y``
          - ``CONTAINS``, equating to ``x in y`` """

    from canteen import model
    value = model.AbstractModel._PropertyValue(value)  # make a value

    # repeated properties do not support EQUALS -> only CONTAINS
    if prop and prop._repeated and operator is EQUALS:
      self.operator = operator = CONTAINS

    self.target, self.value, self.kind, self.operator, self.chain = (
      prop, value, type, operator, [])

    if AND or OR:
      for var in (AND, OR):
        self.chain += (var if (
          isinstance(var, (list, tuple, set, frozenset))) else [var])

  def __repr__(self):

    """ Generate a string representation of this :py:class:`Filter`.

        :returns: """

    return 'Filter(%s %s %s)' % (
      (self.sub_operator.name + ' ') if self.sub_operator else (
        '' + getattr(self.target, 'name', self.generic_name)),
      _operator_strings[self.operator], str(self.value))

  def AND(self, filter_expression):

    """ Chain a query with this one, logically separated by an ``AND`` operator.

        :param filter_expression: Other filter to add to this filter's chain.

        :returns: ``self``, for the ability to further chain this query."""

    self.chain += [filter_expression.set_subquery_operator(AND)]
    return self

  def OR(self, filter_expression):

    """ Chain a query with this one, logically separated by an ``OR`` operator.

        :param filter_expression: Other filter to add to this filter's chain.

        :returns: ``self``, for the ability to further chain this query."""

    self.chain += [filter_expression.set_subquery_operator(OR)]
    return self

  def set_subquery_operator(self, operator):

    """ Internal function for setting a query object's subquery operator, which
        is required to support ``AND/OR`` filter chaining.

        :param operator: Set the local query's subquery operator to this value.

        :returns: ``self``, for the ability to further chain this query. """

    self.sub_operator = operator
    return self

  def match(self, target):

    """ Match this query's target, operator, and embedded data against a target
        entity or value.

        :param target: May be a full ``Model``  object or raw value. Matched
          against this handler's constraints.

        :raises:

        :returns: ``True`` if the target ``Model`` or value matches this
          filter's contraints, ``False`` otherwise. """

    if self.operator not in _operator_map:  # pragma: no cover
      raise RuntimeError('Invalid comparison operator'
                         ' could not be matched: "%s".' % self.operator)

    try:
      # it's a raw entity -> dict
      if isinstance(target, dict):
        if self.target.name in target:
          return _operator_map[self.operator](*(
            target[self.target.name], self.value.data))
        # no such property
        return False  # pragma: no cover

      # it's a raw value of some sort
      if isinstance(target, self.target._basetype):
        return _operator_map[self.operator](target, self.value.data)

      # it's a model or something
      return _operator_map[self.operator](*(
        getattr(target, self.target.name), self.value.data))

    except AttributeError:  # pragma: no cover
      return False  # no such property

  def __call__(self, target):

    """ Proxy to ``self.match``, which matches a target ``Model`` or value
        against this filter's constraints.

        :param target: Target ``model.Model`` instance or raw value to match
          against.

        :returns: ``True`` if ``target`` matches, otherwise ``False``. """

    return self.match(target)  # pragma: no cover


class KeyFilter(Filter):

  """ Expresses a filter that applies to an entity's associated
      :py:class:`model.Key`, or one of the member components thereof. """

  # == Constants == #
  KIND = KEY_KIND
  ANCESTOR = KEY_ANCESTOR

  def __init__(self, value,
                      AND=None,
                      OR=None,
                      type=KEY_KIND):

    """ Initialize this ``KeyFilter`` with either ``KIND`` or ``ANCESTOR``
        filter modes and a ``value`` to filter against.

        :param value: Value to filter entity keys against.

        :param AND: Subfilters to chain with a logical ``AND``.

        :param OR: Subfilters to chain with a logical ``OR``.

        :param type: Type of filter to create, defaults to ``KIND``, meaning it
          is a filter against a key's kind name. The other option is
          ``ANCESTRY``, which filters against a key's ancetry path. """

    super(KeyFilter, self).__init__(None, value, AND=AND, OR=OR, type=type)

  @property
  def generic_name(self):  # pragma: no cover

    """ Return a generic string name for labels including this ``KeyFilter``.

        :returns: Static string label, depending on filter type. """

    return 'Key' if self.kind is self.KIND else 'Ancestor'


class EdgeFilter(Filter):

  """ Expresses a filter that applies to a ``Vertex`` objects' undirected or
      directed ``Edge`` objects. """

  # == Constants == #
  EDGES = EDGES
  NEIGHBORS = NEIGHBORS

  # == Variables == #
  tails = None

  def __init__(self, value,
                      tails=None,
                      AND=None,
                      OR=None,
                      type=EDGES):

    """ Initialize this ``EdgeFilter`` with either ``EDGES`` or ``NEIGHBORS``
        filter modes and a ``value`` to filter against.

        :param value: Value to filter relationships against.

        :param tails: If this is a query for directed edges, this should be
          ``True`` or ``False`` (for whether *heads* or *tails* are desired).
          Otherwise, if it is ``None`` (the default), this implies a query for
          undirected edges.

        :param AND: Subfilters to chain with a logical ``AND``.

        :param OR: Subfilters to chain with a logical ``OR``.

        :param type: Type of filter to create, defaults to ``KIND``, meaning it
          is a filter against a key's kind name. The other option is
          ``ANCESTRY``, which filters against a key's ancetry path. """

    self.tails = tails
    super(EdgeFilter, self).__init__(None, value, type=type, operator=CONTAINS)

  @property
  def generic_name(self):

    """ Return a generic string name for labels including this ``EdgeFilter``.

        :returns: Static string label, depending on filter type. """

    return 'Edges' if self.kind is self.EDGES else 'Neighbors'



class Sort(QueryComponent):

  """ Expresses a directive to sort resulting entities by a property in a given
      direction. """

  ## == Sort Orders == ##
  ASC = ASCENDING = ASC
  DSC = DESCENDING = DSC

  def __init__(self, prop, type=PROPERTY, operator=ASC):

    """ Initialize this :py:class:`Sort`.

        :param prop: Subject property to sort upon.

        :param type: Type of sort to apply, defaults to a ``PROPERTY``-based
          sort, which sorts upon property values.

        :param operator: Either ascending (``ASC`` or ``ASCENDING`` global
          symbols) or descending (``DSC`` or ``DESCENDING`` global symbols). """

    self.target, self.kind, self.operator = prop, type, operator

  def __repr__(self):

    """ Generate a string representation of this :py:class:`Sort`.

        :returns: Pleasant string representation of this ``Sort``, formatted
          like ``Sort(target, operator)``. """

    return 'Sort(%s, %s)' % (self.target.name, self.operator)



# -*- coding: utf-8 -*-

"""

  logic base
  ~~~~~~~~~~

  Provides base configuration and code to setup Canteen ``Logic`` classes, which
  allow a developer to provide a piece of cross-cutting functionality to their
  entire application.

  ``Logic`` classes are Canteen ``Component``s, meaning they provide injectable
  structures for the DI engine. ``Logic`` classes, when bound to a string name
  with ``decorators.bind``, will be accessible from ``Compound`` object methods
  at ``self.name``.

  Example:

    # -*- coding: utf-8 -*-
    from canteen import decorators, Logic

    @decorators.bind('math')
    class Math(Logic):


  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# core API
from ..core import meta
from ..util import decorators


@decorators.singleton
class Logic(object):

  """ Base class for Canteen ``Logic`` components. Specifies a class tree on the
      ``Proxy.Component`` side (meaning that it *provides* DI resources, instead
      of a ``Proxy.Compound``, which *consumes* them).

      ``Logic`` classes must be bound using the ``decorators.bind`` tool, which
      takes a ``str`` name and binds a ``Logic`` class (or child) to that name
      on all ``Compound`` classes and objects. """

  __owner__, __metaclass__ = "Logic", meta.Proxy.Component



# -*- coding: utf-8 -*-

"""

  base protocol tests
  ~~~~~~~~~~~~~~~~~~~

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# testing
from canteen import test

# base protocol
from canteen.base import protocol


class SomeValidProtocol(protocol.Protocol):

  """ I am a valid, registered protocol """

  def encode_message(self, message):

    """ sample encode message """

  def decode_message(self, _type, encoded):

    """ sample decode message """


class BaseProtocolTest(test.FrameworkTest):

  """ Tests `base.protocol`. """

  def _make_protocol(self, valid=True):

    """ Build a quick mock Protocol """


    if valid:


      class _SomeValidProtocol(protocol.Protocol):

        """ I am a valid, registered protocol """

        def encode_message(self, message):

          """ sample encode message """

        def decode_message(self, _type, encoded):

          """ sample decode message """

      return _SomeValidProtocol

    else:


      class SomeInvalidProtocol(protocol.Protocol):

        """ I am an invalid protocol """


      return SomeInvalidProtocol

  def test_base_protocol(self):

    """ Test that `base` exports `Protocol` """

    assert hasattr(protocol, 'Protocol')

  def test_protocol_abstract(self):

    """ Test that `Protocol` is abstract """

    with self.assertRaises(TypeError):
      self._make_protocol(valid=False)()

  def test_protocol_extend(self):

    """ Test that `Protocol` can be extended """

    self._make_protocol()()

  def test_protocol_register(self):

    """ Test that `Protocol` registers properly """

    # simulate decorator
    protocol.Protocol.register('randorpc', (
      'application/rando',
      'application/x-rando'
    ))(SomeValidProtocol)

  def test_protocol_all(self):

    """ Test that `Protocol.all` iterates over registered protocols """

    # simulate decorator
    protocol.Protocol.register('randorpc', (
      'application/rando',
      'application/x-rando'
    ))(SomeValidProtocol)

    all_protocols = [p for p in protocol.Protocol.all]
    assert SomeValidProtocol in all_protocols

  def test_protocol_mapping(self):

    """ Test that `Protocol.mapping` returns a proper type=>protocol
        mapping """

    # simulate decorator
    protocol.Protocol.register('randorpc', (
      'application/rando',
      'application/x-rando'
    ))(SomeValidProtocol)

    map = protocol.Protocol.mapping

    assert isinstance(*(
      map.lookup_by_name('randorpc').protocol, SomeValidProtocol))

    assert isinstance(*(
      map.lookup_by_content_type('application/rando').protocol,
      SomeValidProtocol))



# -*- coding: utf-8 -*-

"""

  datastructures
  ~~~~~~~~~~~~~~

  lightweight datastructures for use inside and outside
  :py:class:`canteen`.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# stdlib
import abc

# canteen util
from . import decorators


## Globals
_ENUM_TYPES = int, long, float, basestring, bool


class Sentinel(object):

  """ Create a named sentinel object. """

  name, hash, _falsy = None, None, False

  def __init__(self, name, falsy=False):

    """ Construct a new ``Sentinel``, which is essentially just a symbolic
        object at a simple string name. Two ``Sentinel``s with the same name
        evaluate to be *equal* to each other.

        :param name: Simple string name for the new ``Sentinel``-to-be.

        :param falsy: Whether the resulting ``Sentinel`` object should evaluate
          as *falsy* (if it is to stand-in for ``None`` or ``False``, for
          instance). """

    self.name, self.hash, self._falsy = (
      name, int((''.join(str(ord(c)) for c in name))), falsy)

  # hash value for this sentinel
  __hash__ = lambda self: self.hash

  # equality comparator
  __eq__ = lambda self, other: (
    isinstance(other, self.__class__) and other.hash == self.hash)

  # string representation
  __repr__ = lambda self: '<Sentinel "%s">' % self.name

  # stringification value
  __str__ = __unicode__ = lambda self: self.name

  # falsyness
  __nonzero__ = lambda self: (not self._falsy)


# Sentinels
EMPTY, _TOMBSTONE = (
  Sentinel("EMPTY", True),
  Sentinel("TOMBSTONE", True))


class UtilStruct(object):

  """ Abstract class for a utility object. """

  __metaclass__ = abc.ABCMeta

  ## Init -- Accept structure fill
  def __new__(cls, *args, **kwargs):

    """ Class constructor that enforces abstractness at the root of the class
        tree.

        :param *args:
        :param **kwargs:

        :raises NotImplementedError: If the root class :py:class:`UtilStruct` is
          constructed directly. Otherwise, returns a new instance of the
          requested class. """

    if cls.__name__ is 'UtilStruct':
      raise NotImplementedError('Cannot construct `UtilStruct` directly as'
                                ' it is an abstract class.')
    return object.__new__(cls, *args, **kwargs)

  @abc.abstractmethod
  def fillStructure(self, struct, case_sensitive=False, **kwargs):

    """ Abstract method that fills a local object with data, usually from
        initialization.

        :param struct:
        :param case_sensitive:
        :param kwargs:

        :raises:
        :returns: """

    raise NotImplementedError('`UtilStruct.fillStructure` is abstract and must'
                              ' be implemented by a subclass.')


class ObjectProxy(UtilStruct):

  """ Same handy object as above, but stores the entries in an _entries
      attribute rather than the class dict.  """

  _entries = _case_sensitive = None

  def __init__(self, struct=None, case_sensitive=False, **kwargs):

    """ If handed a dictionary (or something) in init, send it to fillStructure
        (and do the same for kwargs).

        :param struct:
        :param case_sensitive:
        :param kwargs:

        :raises TypeError:

        :returns: """

    self._entries, self._case_sensitive = {}, case_sensitive
    if struct:
      if kwargs: struct.update(kwargs)
      self.fillStructure(struct, case_sensitive=case_sensitive)

  def fillStructure(self, fill, case_sensitive=False, **kwargs):

    """ If handed a dictionary, will fill self with those entries. Usually
        called from ``__init__``.

        :param fill: Structure to fill self with.

        :param case_sensitive: Whether we should initialize while ignoring case.

        :param kwargs: Keyword arguments to be applied to ``struct`` as
          override.

        :returns: ``self``. """

    self._case_sensitive = case_sensitive
    if fill:
      if kwargs: fill.update(kwargs)
      for k, v in (fill.iteritems() if isinstance(fill, dict) else iter(fill)):
        self._entries[self.i_filter(k)] = v
    return self

  def __getitem__(self, name):

    """ 'x = struct[name]' override.

        :param name:
        :raises KeyError:
        :returns: """

    filtered = self.i_filter(name)
    if filtered not in self._entries and name not in self._entries:
      raise KeyError("Cannot locate name '%s'"
                     " in ObjectProxy '%s'." % (name, self))
    return self._entries.get(filtered, self._entries.get(name))

  def __getattr__(self, name):

    """ 'x = struct.name' override.

        :param name:
        :raises AttributeError
        :returns: """

    filtered = self.i_filter(name)
    if filtered not in self._entries and name not in self._entries:
      raise AttributeError("Could not find the attribute '%s'"
                           " on the specified ObjectProxy." % name)
    return self._entries.get(filtered, self._entries.get(name))

  # filter for case sensitivity
  i_filter = lambda self, target: (
    (self._case_sensitive and target) or str(target).lower())

  # contains override
  __contains__ = contains = lambda self, name: (
    self.i_filter(name) in self._entries)

  # dict-style buffered access
  keys = lambda self: self._entries.keys()
  values = lambda self: self._entries.values()
  items = lambda self: self._entries.items()

  # dict-style streaming access
  iterkeys = lambda self: self._entries.iterkeys()
  itervalues = lambda self: self._entries.itervalues()
  iteritems = lambda self: self._entries.iteritems()


class WritableObjectProxy(ObjectProxy):

  """ Same handy object as `ObjectProxy`, but allows appending things at
      runtime. """

  def __setitem__(self, name, value):

    """ 'struct[name] = x' override.

        :param name:
        :param value:
        :returns: """

    self._entries[name] = value

  def __setattr__(self, name, value):

    """ 'struct.name = x' override.

        :param name:
        :param value:
        :returns: """

    if name in ('_entries', '_case_sensitive', '__slots__'):
      return object.__setattr__(self, name, value)
    self._entries[name] = value

  def __delattr__(self, name):

    """ 'del struct.name' override.

        :param name:
        :raises AttributeError:
        :returns: """

    if self.i_filter(name) not in self._entries:
      raise AttributeError("Could not find the entry '%s'"
                           " on the specified ObjectProxy." % name)
    del self._entries[self.i_filter(name)]

  def __delitem__(self, name):

    """ 'del struct[name]' override.

        :param name:
        :raises KeyError:
        :returns: """

    if self.i_filter(name) not in self._entries:
      raise KeyError("Could not find the entry '%s'"
                     " on the specified ObjectProxy." % name)
    del self._entries[self.i_filter(name)]


class CallbackProxy(ObjectProxy):

  """ Handy little object that takes a dict and makes it accessible via
      var[item], but returns the result of an invoked ``callback(item)``. """

  _entries = None  # cached entries
  callback = None  # callback func

  # noinspection PyMissingConstructor
  def __init__(self, callback, struct=None, **kwargs):

    """ Map the callback and fillStructure if we get one via `struct`.

        :param callback: Callable to produce return values, given the ``item``
          or ``attr`` desired. If this ``CallbackProxy`` has a set of registered
          ``self._entries``, and the requested ``item`` or ``attr`` is valid,
          the value at ``item`` or ``attr`` in ``self._entries`` will be passed
          as the first and only argument to ``callback`` when looking for a
          return value. If this ``CallbackProxy`` is unregistered, the desired
          ``item`` or ``attr`` name is passed as the first and only argument to
          ``callback``.

        :param struct: Structure (``dict``) of items to register in this
          ``CallbackProxy``'s local ``self._entries`` registry. If this
          structure is provided, it will be validated against for ``item`` or
          ``attr`` requests, and will be handed to the ``callback`` function
          as described above.

        :param **kwargs: ``key=value`` pairs to override in ``struct`` before
          registering in ``self._entries``. """

    struct = struct or {}
    self.callback = callback

    self._entries = struct
    if kwargs: self._entries.update(kwargs)

  def __getitem__(self, item):

    """ 'x = struct[item]' override.

        :param item: Item to fetch via ``self.callback``. If there are
          registered entries on the local ``CallbackProxy``, ``item`` must be
          present as a key, and the value at ``item`` is passed to the local
          ``self.callback`` for resolving a value.

        :raises KeyError: Raised if this ``CallbackProxy`` has registered
          ``self._entries`` and ``item`` is not present as a key. ``KeyError``s
          raised from ``self.callback`` are also bubbled.

        :returns: Result of ``self.callback(self._entries[item])`` if this
          ``CallbackProxy`` has registered ``self._entries``, or the result of
           ``self.callback(item)`` if there are no registered
           ``self._entries``. """

    if self._entries and item not in self._entries:
      raise KeyError("Could not retrieve item '%s'"
                     " from CallbackProxy '%s'." % (item, self))
    return self.callback((self._entries.get(item) if self._entries else item))

  def __getattr__(self, attr):

    """ 'x = struct.attr' override.

        :param attr: String attribute name to fetch from this ``CallbackProxy``,
          either via the result of ``self.callback(self._entries[attr])`` if
          there are registered ``self._entries`` or otherwise the result of
          ``self.callback(attr)``.

        :raises AttributeError: Raised if this ``CallbackProxy`` has registered
          ``self._entries`` and ``item`` is not present as a key.
          ``AttributeError``s raised from ``self.callback`` are also bubbled.

        :returns: Result of the ``callback`` dispatch as described above. """

    if self._entries and attr not in self._entries:
      raise AttributeError("CallbackProxy could not resolve entry '%s'." % attr)
    return self.callback((self._entries.get(attr) if self._entries else attr))

  def __contains__(self, item):

    """ 'x in struct' override.

       :param item: Item to check the local container for.

       :return: Whether or not ``item`` is held by ``self._entries``, or
         ``False`` if it could not be found (or if this ``CallbackProxy`` has no
         registered ``self._entries`` table). """

    return self._entries and item in self._entries

  # `struct()` override
  __call__ = lambda self, *args, **kwargs: self.callback(*args, **kwargs)


class BidirectionalEnum(object):

  """ Small and simple datastructure for mapping static names to symbolic
      values. Interoperable with the RPC and model layers. """

  __singleton__, __slots__ = True, tuple()

  class __metaclass__(type):

    """ Metaclass for property-gather-enabled classes. """

    def __new__(mcs, name, chain, _map):

      """ Read mapped properties, store on the object, along with a reverse
          mapping.

          :param name: Name of the ``BidirectionalEnum`` subtype to factory.
          :param chain: Inheritance chain for target subtype.
          :param _map: ``dict`` map of ``key = value`` attribute pairs.

          :raises RuntimeError: If a non-unique ``key`` or ``value`` is given
            as part of the ``BidirectionalEnum``-to-be's ``map``.

          :returns: Constructed ``BidirectionalEnum`` subclass. """

      # init calculated data attributes
      _pmap, _keys, _plookup = (
          _map['_pmap'], _map['_keys'], _map['_plookup']) = (
            {}, [], set())

      # set class-level enum properties
      for key, value in _map.iteritems():
        if not key.startswith('_') and isinstance(value, _ENUM_TYPES):
          if value in _plookup:  # pragma: no cover
            raise RuntimeError('Cannot map non-unique (double-validated)'
                               ' `BidirectionalEnum` value %s.' % value)
          if key in _plookup:  # pragma: no cover
            raise RuntimeError('Cannot map non-unique (double-validated)'
                               ' `BidirectionalEnum` key %s.' % value)

          # things look good: add the value
          _keys.append(key)
          _pmap[key] = value
          _pmap[value] = key
          _plookup.add(key)
          _plookup.add(value)

      _map['_keys'] = tuple(_keys)
      return type.__new__(mcs, name, chain, _map)

    def __iter__(cls):

      """ Iterate over all enumerated values.

         :returns: ``(k, v)`` tuples for each enumerated item and value, one at
          a time. """

      for k in cls._keys:
        yield k, getattr(cls, k)

    def __getitem__(cls, item):

      """ Fetch an item from the local ``BidirectionalEnum`` (forward resolve).

          :raises KeyError: If ``item`` is not a valid item registered on the
            local ``BidirectionalEnum``.

          :returns: Bound value at the (forward-resolved) item on the local
            ``BidirectionalEnum``. """

      return cls._pmap[item]

    def __contains__(cls, item):

      """ Check if an ``item`` is a registered member or value of the local
          ``BidirectionalEnum``.

          :param item: Item to check for membership against the local enum.

          :return: ``bool`` - ``True`` if ``item`` is a valid member with a
            bound value retrievable via item or attribute syntax or a value with
            a bound key retreivable via ``reverse_resolve``, otherwise
            ``False``. """

      return item in cls._plookup

    def __setitem__(cls, item, value):

      """ Disallow writing to items registered in the ``BidirectionalEnum``,
          as it is an immutable type.

          :param item: Item to write.
          :param value: Value to write at ``item``.

          :raises NotImplementedError: Always, as ``BidirectionalEnum`` is an
            immutable type. """

      raise NotImplementedError('`%s` is an immutable type'
                                ' and cannot be written to with item'
                                ' syntax. ' % cls.__name__)

    def __setattr__(cls, attr, value):

      """ Disallow writing to attributes registered in the ``BidirectionalEnum``,
          as it is an immutable type.

          :param attr: Attribute to write.
          :param value: Value to write at ``attr``.

          :raises NotImplementedError: Always, as ``BidirectionalEnum`` is an
            immutable type. """

      raise NotImplementedError('`%s` is an immutable type'
                                ' and cannot be written to with attribute'
                                ' syntax. ' % cls.__name__)

  def __new__(cls):

    """ Disallow direct construction of ``BidirectionalEnum`` objects.

        :raises TypeError: ``BidirectionalEnum`` is both an abstract class
          (itself) and an object that is always used as its type. Thus, objects
          of ``BidirectionalEnum`` are not allowed to be created and attempts
          to do so always result in ``TypeError``s. """

    raise TypeError('`%s` objects are abstract'
                    ' and only usable as type objects.' % cls.__name__)

  # forward and reverse resolve
  reverse_resolve = (
    classmethod(lambda cls, code: cls._pmap.get(code, False)))
  forward_resolve = resolve = (
    classmethod(lambda cls, flag: cls.__getattr__(flag, False)))

  # serialization and string repr
  __json__ = classmethod(lambda cls: cls.__serialize__())
  __serialize__ = (
    classmethod(lambda cls: dict(((k, v) for k, v in cls._plookup if not (
      k.startswith('_'))))))

  __repr__ = classmethod(lambda cls: ('::'.join([
      "<%s" % cls.__name__,
      ','.join([
        block for block in (
          '='.join([str(k), str(v)]) for k, v in cls.__serialize__().items())]),
      "BiDirectional>"])))



# -*- coding: utf-8 -*-

"""

  logic
  ~~~~~

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# submodules
from .http import *
from .cache import *
from .assets import *
from .session import *
from .template import *
from .realtime import *


__all__ = ('url',
           'http',
           'cache',
           'realtime',
           'template')



# -*- coding: utf-8 -*-

"""

  runtime
  ~~~~~~~

  holds code that bridges :py:mod:`canteen` into various WSGI runtimes,
  like :py:mod:`gevent` and :py:mod:`wsgiref`.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# runtimes
from . import uwsgi
from . import wsgiref
from . import werkzeug


__all__ = ('uwsgi',
           'wsgiref',
           'werkzeug')



# -*- coding: utf-8 -*-

"""

  caching logic
  ~~~~~~~~~~~~~

  exposes simple logic for caching objects in-memory or in
  caching engines like ``memcached``.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# stdlib
import abc
import time
import weakref
import threading

# core and utils
from ..base import logic
from ..util import decorators


## Globals
_caches = {}
_default = (threading.local(), None)
_BUILTIN_TYPES = (int, float, str, list, dict, tuple, unicode, type(abc))


class Cache(object):

  """ Specifies an instance of a cache, that supports storage and retrieval of
      arbitrary native Python values.

      Attached to every :py:class:`Cache` is a ``Strategy`` and a ``Target``,
      specifying the management policy and underlying caching engine,
      respectively.

      Both components are extendable to create custom caching solutions. """

  class Engine(object):

    """  """

    target = None  # cache adapter/library
    strategy = None  # strategy to use for eviction

    __metaclass__ = abc.ABCMeta

    ## == Internals == ##
    def __init__(self, target, strategy=None):

      """ Initialize a new ``Cache``, utilizing ``target`` as a cache
          adapter/library ``strategy`` as a cache management policy.

          :param target: Target library or cache system.

          :param strategy: Cache management policy, extending
            ``Cache.Strategy``. """

      self.target, self.strategy = target, strategy

    #### ==== Read Methods ==== ####
    @abc.abstractmethod
    def get(self, key):

      """ Specifies abstract version of ``Cache.get``, which is used to retrieve
          cached items.

          :param key: Cached item's key.

          :raises NotImplementedError: Always, as this method is abstract.

          :returns: Implementors must return the cached item stored at
            ``key``. """

      raise NotImplementedError('`Cache.get`'
                                ' is abstract.')  # pragma: no cover

    @abc.abstractmethod
    def get_multi(self, keys):

      """ Specifies abstract version of ``Cache.get_multi`, which is used to
          retrieve multiple cache items ine one call.

          :param keys: Iterable of cached item keys.

          :raises NotImplementedError: Always, as this method is abstract.

          :returns: Implementors must return a dict of cached items cached
            items, or ``None``, according to that result's position in the
            original ``keys`` iterable. """

      raise NotImplementedError('`Cache.get_multi`'
                                ' is abstract.')  # pragma: no cover

    @abc.abstractmethod
    def items(self, timestamp):

      """ Specifies abstract version of ``Cache.items``, which is used to
          iterate over items in the cache, optionally current to ``timestamp``.

          This method is optional for implementors but must always be defined.
          If an implementing class wishes not to implement this method, it must
          return ``NotImplemented``.

          :param timestamp: Optionally only return cached items that were saved
            after the given ``datetime.datetime`` or integer Unix timestamp.

          :raises NotImplementedError: Always, as this method is abstract.

          :returns: Implementors must either return ``NotImplemented`` (if they
            choose not to provide an ``items`` implementation) or an iterator
            that will yield ``(key, value)`` tuples of items in the cache,
            optionally current to ``timestamp``. """

      raise NotImplementedError('`Cache.items`'
                                ' is abstract.')  # pragma: no cover


    #### ==== Write Methods ==== ####
    @abc.abstractmethod
    def set(self, key, value):

      """ Specifies abstract version of ``Cache.set``, which is used to store
          individual items in the ``Cache``.

          :param key: String key the item should be stored at, for future
            reference.

          :param value: Raw value to be stored in the cache at ``key``.

          :raises NotImplementedError: Always, as this method is abstract.

          :returns: Implementors must return the original value passed in to be
            cached at ``value``. """

      raise NotImplementedError('`Cache.set`'
                                ' is abstract.')  # pragma: no cover

    @abc.abstractmethod
    def set_multi(self, map):

      """ Specifies abstract version of ``Cache.set_multi``, which is used to
          write multiple items to the ``Cache`` in one call.

          :param map: Iterable of ``(key, value)`` tuples or ``dict`` of
            ``key=>value`` mappings that should be saved to the cache.

          :raises NotImplementedError: Always, as this method is abstract.

          :returns: Implementors must return the original iterable or ``dict``
            of values passed in to be cached at ``map``. """

      raise NotImplementedError('`Cache.set_multi`'
                                ' is abstract.')  # pragma: no cover

    #### ==== Delete Methods ==== ####
    @abc.abstractmethod
    def delete(self, key):

      """ Specifies abstract version of ``Cache.delete``, which is used to
          delete an individual item from the ``Cache``.

          :param key: String key under which an item may be stored in the
            ``Cache`` that should be deleted.

          :raises NotImplementedError: Always, as this method is abstract.

          :returns: Implementors must return the original ``key`` at which an
            item should have been deleted. """

      raise NotImplementedError('`Cache.delete`'
                                ' is abstract.')  # pragma: no cover

    @abc.abstractmethod
    def delete_multi(self, keys):

      """ Specifies abstract version of ``Cache.delete_multi``, which is used to
          delete multiple items from the ``Cache`` in one call.

          :param keys: Iterable of string keys under which existing cached items
            should be deleted.

          :raises NotImplementedError: Always, as this method is abstract.

          :returns: Implementors must return the original iterable of ``keys``
            under which items should have been deleted. """

      raise NotImplementedError('`Cache.delete_multi`'
                                ' is abstract.')  # pragma: no cover

    @abc.abstractmethod
    def clear(self):

      """ Specifies abstract version of ``Cache.clear``, which is used to nuke
          the cache (clearing it of all items) in one call.

          :returns: ``None``, as the cache should have been nuked and what else
            would we return? """

      raise NotImplementedError('`Cache.clear`'
                                ' is abstract.')  # pragma: no cover

  class Strategy(object):

    __metaclass__ = abc.ABCMeta

    #### ==== Expiration Methods ==== ####
    @abc.abstractmethod
    def should_expire(self, key):

      """ Specifies abstract version of ``Cache.Strategy.should_expire``, which
          is called by the ``Cache`` internals on an implementing cache strategy
          to discern whether a key should be dropped from the cache.

          :param key: String key that should be evaluated against this cache
            management policy.

          :returns: Implementors are expected to return ``True`` if the key
            should expire or ``False`` if the key should be kept. """

      raise NotImplementedError('`Cache.Strategy.should_expire`'
                                ' is abstract.')  # pragma: no cover

    #### ==== Scan Methods ==== ####
    @abc.abstractmethod
    def tick(self, timestamp):

      """ Specifies abstract version of ``Cache.Strategy.tick``, which is called
          by the ``Cache`` internals every so often to trim or clean the cache.

          :param timestamp: Current integer Unix timestamp, handed-in for
            convenience.

          :returns: Nothing is expected of implementors for a return value from
            this method. """

      raise NotImplementedError('`Cache.Strategy.tick`'
                                ' is abstract.')  # pragma: no cover


@decorators.bind('cache')
class Caching(logic.Logic):

  """ Bundled logic that provides basic caching functionality through Canteen's
      builtin ``Cache`` APIs. Formerly the 'Core Cache API'. """

  #### ==== Cache Strategies ==== ####
  class PersistentCache(Cache.Strategy):

    """ Offers a ``Cache.Strategy`` designed to *never* expire keys and refuse
        to clean/trim during cache ``tick``. """

    def should_expire(self, key, timestamp):

      """ Always return ``False`` as this strategy specifies that no keys should
          be dropped, ever.

          :param key: String ``key`` to not-drop.

          :param timestamp: Integer ``timestamp`` for ``now``, as understood by
            the client. Handed in for convenience.

          :returns: ``False``. Always. """

      return False  # nope

    def tick(self, timestamp):

      """ Always return ``NotImplemented`` as a sentinel to the ``Cache``
          internals that this strategy refuses to clean/trim on cache tick.

          :returns: ``NotImplemented``. Always. """

      return NotImplemented  # double nope


  #### ==== Cache Engines ==== ####
  class Threadcache(Cache.Engine):

    """ Manages a simple thread local-backed caching engine, suitable for
        caching basic items that don't relate to HTTP state. """

    def get(self, key, default=None, _skip=False):

      """ Retrieve an item from the threadcache by key.

          :param key: Key at which an item should exist in the threadcache to
            return to the callee.

          :param default: Default item to return in the case of a nonexistent
            key. Defaults to ``None``.

          :param _skip: Internal flag that indicates we can skip checking
            ``key``'s existence, because somehow we already know it exists.

          :raises KeyError: In the case of a nonexistent ``key``, but a truthy
            value for ``_skip``, as this exposes the raw access without a
            check through ``__contains__``.

          :returns: The item cached at ``key``, if any, or ``default``. """

      if _skip or key in self.target:

        # retrieve
        value, timestamp = self.target[key]

        # dereference weakref
        if isinstance(value, weakref.ref):  # pragma: no cover
          value = value()

        # check expiration and ref and return
        if value is not None:
          if not self.strategy.should_expire(key, timestamp):
            return value
          else:  # pragma: no cover
            self.delete(key)
      return default

    def get_multi(self, keys, default=None):

      """ Retrieve multiple items at once from the threadcache by key. Basically
          a batch version of ``get``.

          :param keys: Iterable of keys to retrieve from the threadcache via
            ``get``.

          :param default: Default value to pack the resultlist with for
            nonexistent keys.

          :returns: Iterable of results for each key in ``keys``, or ``default``
            in place of items that don't exist. Order is maintained for the
            provided ``keys`` iterable. """

      return dict(((key, self.get(key, default)) for key in keys))

    def items(self):

      """ Iterate over all available keys in the cache, one by one, yielding
          them as ``(key, value)`` pairs.

          :yields: Two-item tuples of ``(key, value)`` pairs, for each ``key``
            accessible in the local ``Threadcache`` at ``self.target``. """

      for key in self.target:
        yield key, self.get(key, _skip=True)

    def set(self, key, value):

      """ Set a value in the threadcache by key.

          :param key: Key at which the value should be cached and potentially
            retrievable later.

          :param value: Value that should be stored in the cache under ``key``.
            Can be any Python-native value or value easily reducible to a raw
            Python-native value.

          :returns: Value stored in the cache. """

      value = (
        weakref.ref(value) if not isinstance(value, _BUILTIN_TYPES) else (
          value))

      self.target[key] = (value, time.time())
      return value

    def set_multi(self, map):

      """ Batch version of ``set``, defined above. Accepts a ``map`` of ``keys``
          to ``values`` which should be persisted via ``set``.

          :param map: ``dict`` or iterable of ``(key, value)`` pairs that should
            be stored in the cache.

          :returns: The fully-buffered set of items stored in the cache. """

      for key, value in (map.iteritems() if isinstance(map, dict) else map):
        self.set(key, value)
      return map

    def delete(self, key):

      """ Delete an item in the cache by ``key``.

          :param key: Key at which any matching values should be removed from
            the threadcache. """

      if key in self.target:
        del self.target[key]

    def delete_multi(self, keys):

      """ Batch version of ``delete``. Accepts an iterable of ``keys`` to delete
          from the local threadcache.

          :param keys: Iterable of keys to be deleted via ``delete``. """

      for key in keys:
        self.delete(key)

    def clear(self):

      """ Clear the entire threadcache of all values in one go.

          :returns: Number of items cleared from the threadcache. """

      length = len(self.target)
      self.target = {}
      return length

  #### ==== Internals ==== ####
  @property
  def config(self):

    """ Property accessor for caching configuration.

        :returns: Any application or framework configuration at the path
          ``Caching``. """

    from canteen.util import config  # pragma: no cover
    return config.Config().config.get('Caching', {})  # pragma: no cover

  @property
  def debug(self):  # pragma: no cover

    """ Property accessor for debug mode.

        :returns: Boolean indicating whether ``debug`` mode should be active.
          Defaults to Python's internal ``__debug__`` state. """

    return self.config.get('debug', __debug__)  # pragma: no cover

  #### ==== Cache Management ==== ####
  @staticmethod
  def spawn(name=None,
            target=None,
            engine=Threadcache,
            strategy=PersistentCache):

    """ Spawn a new cache, at name ``name``, target ``target``, and optionally
        with an ``engine``/``strategy`` pair.

        Calling ``spawn`` without any arguments returns a reference to the
        default (thread-local) ``Threadcache``, used for general state storage
        for internal operations like routing.

        :param name: Simple string name that can be used to refer to this
          particular cache. Must remain unique across caches.

        :param target: Target transport or underlying storage, if applicable.
          This is an implementation-dependent parameter, and could, for example,
          be a prebuilt memcache client object, a global variable, or a hot
          Redis connection.

        :param engine: Caching engine to use against ``target``. The engine is
          in charge of knowing how to ``set``/``get``/``delete`` items from a
          kind of caching system, such as inmemory caching or ``memcache``.

        :param strategy: Caching strategy use for the management of memory
          pressure. Default strategy is ``PersistentCache``, meaning items are
          kept for the entire lifetime of the local thread. ``Cache.Strategy``
          classes can specify methods to hook into the normal operation of
          canteen's runtime to perform items on ``tick`` or for expiration
          checks.

        :returns: Reference to the newly-spawned ``Cache`` object, prepared and
          registered with the desired ``target``/``engine``/``strategy``.  """

    global _caches
    global _default

    _localtarget, cache = _default
    if not name:
      if not cache:
        _default = _caches['__default__'] = (
          _localtarget, engine(target=_localtarget.__dict__,
                               strategy=strategy()))
      return _default[1]

    _caches[name] = engine(target=target or threading.local().__dict__,
                           strategy=strategy())
    return _caches[name]

  @staticmethod
  def clear(name=None):

    """ Clear the entire contents of a ``name``d cache. If no name is provided,
        will clear contents of *all known* caches.

        :param name: Name of the cache to clear all items from. Defaults to
          ``None``, which clears the contents of *all known* caches.

        :returns: A tuple, consisting of ``num_caches_cleared`` (a count of
          individual ``Cache`` objects cleared) and ``num_items_cleared`` (a
          full count of all items cleared from all ``Cache`` objects), in that
          order.  """

    if not name:
      _total, _cache_count = 0, 0
      for name, cache in _caches.iteritems():
        if name == '__default__':
          _localtarget, cache = cache
        _total += cache.clear()
        _cache_count += 1
      return _cache_count, _total

    if name in _caches:
      return _caches[name].clear()
    return 0, 0  # pragma: no cover

  @staticmethod
  def flush():  # pragma: no cover

    """ Perform a hard flush of the local ``_caches`` index, which should free
        all local ``Threadcache`` items and release all known connections to
        external caches.

        :returns: A tuple of wiped values for module globals ``_caches`` and
          ``_default``. """

    global _caches
    global _default

    _caches, _default = {}, (threading.local(), None)

  @classmethod
  def get(cls, key, default=None):

    """ Retrieve a key from the default cache, at the named location ``key``.

        :param key: Key at which to retrieve a value from the default cache,
          acquired via ``cls.spawn()``.

        :param default: Default value to return in the case of a missing
          ``key``. Defaults to ``None``.

        :returns: Value stored at ``key`` in the default cache, or
          ``default``. """

    return cls.spawn().get(key, default)

  @classmethod
  def get_multi(cls, keys, default=None):

    """ Batch version of ``get``, to retrieve multiple items from the default
        caching engine, acquiref via ``cls.spawn()``.

        :param keys: Iterable of ``key`` locations at which to retrieve values
          from the default ``Threadcache``.

        :param default: Default value to pad resulting list with in case items
          from ``keys`` cannot be resolved.

        :returns: ``list`` of values, synchronized with iterable ``keys`` with
          values from the default threadcache at those ``keys``, or ``default``
          in place of items that could not be found. List order is preserved
          with input.  """

    return dict(((key, cls.spawn().get(key, default)) for key in keys))

  @classmethod
  def items(cls, name=None):

    """ Iterate over all available keys in the named cache at ``name`` (or the
        default local ``Threadcache`` if no ``name`` is provided), one by one,
        yielding them as ``(key, value)`` pairs.

        :yields: Two-item tuples of ``(key, value)`` pairs, for each ``key``
          accessible in the default local ``Threadcache``. """

    for key, value in (
      cls.spawn() if name is None else cls.spawn(name)).items():
      yield key, value

  @classmethod
  def set(cls, key, value):

    """ Set a value in the default ``Threadcache``.

        :param key: Key at which to store ``value`` in the default threadlocal
          ``Threadcache``.

        :param value: Value to store at ``key`` in the default threadlocal
          ``Threadcache``.

        :returns: ``value``, as it was stored in the backing
          ``Threadcache``. """

    return cls.spawn().set(key, value)

  @classmethod
  def set_multi(cls, map):

    """ Batch version of ``set``, to apply a ``map`` of ``(key, value)`` pairs
        to the ``Threadcache``, where each ``value`` is stored at the its
        corresponding ``key`` in the default threadlocal ``Threadcache``.

        :param map: ``dict`` or iterable of ``(key, value)`` tuple pairs to
          store in the default threadlocal ``Threadcache``. """

    _cache = cls.spawn()
    for key, value in (map.iteritems() if isinstance(map, dict) else map):
      _cache.set(key, value)
    return map

  @classmethod
  def delete(cls, key):

    """ Delete an item from the default threadcache at ``key``.

        :param: ``key`` at which any matching value should immediately be
          dropped from the default threadlocal ``Threadcache``.  """

    cls.spawn().delete(key)

  @classmethod
  def delete_multi(cls, keys):

    """ Batch version of ``delete``, to delete multiple `keys`` from the default
        local ``Threadcache`` in one go.

        :param keys: Iterable of keys at which any value should be dropped
          immediately from the default local ``Threadcache`` via ``delete``. """

    _cache = cls.spawn()
    for key in keys:
      _cache.delete(key)



# -*- coding: utf-8 -*-

"""

  core runtime tests
  ~~~~~~~~~~~~~~~~~~

  tests canteen's runtime core, which is responsible for pulling
  the app together with glue and spit.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# stdlib
import operator

# testing
from canteen import test

# runtime core
from canteen.core import runtime


class BaseRuntimeTest(test.FrameworkTest):

  """ Tests :py:mod:`canteen`'s :py:mod:`core.runtime` package,
      which handles pretty much everything from bind to dispatch. """

  def test_exports(self):

    """ Test basic attributes that should always be present on `runtime` """

    assert hasattr(runtime, 'Runtime')
    assert hasattr(runtime, 'Library')


class RuntimeLibraryTest(test.FrameworkTest):

  """ Tests :py:class:`canteen.core.runtime.Library`, which is
      used internally by :py:mod:`canteen` for deferred library
      binding. """

  def test_good_library_string_lax(self):

    """ Test that `Library` can load a known-good module (by path) """

    with runtime.Library('operator') as (library, _operator):
      assert hasattr(_operator, 'eq')
      assert operator.eq is _operator.eq

  def test_good_library_string_strict(self):

    """ Test that `Library` can load a known-good module in `strict` mode """

    with runtime.Library('operator', strict=True) as (library, _operator):
      assert hasattr(_operator, 'eq')
      assert operator.eq is _operator.eq

  def test_bad_library_string_lax(self):

    """ Test that `Library` can ignore a known-bad module (by path) """

    _lib = runtime.Library('i_do_not_exist_at_all_okay_never')
    _lib.__enter__()

  def test_bad_library_string_strict(self):

    """ Test that `Library` can ignore a known-bad module in `strict` mode """

    with self.assertRaises(ImportError):
      _lib = runtime.Library('i_do_not_exist_at_all_okay_never', strict=True)
      _lib.__enter__()

  def test_good_library_module_lax(self):

    """ Test that `Library` can load a module (by mod) """

    with runtime.Library(operator) as (library, _operator):
      assert hasattr(_operator, 'eq')

  def test_good_library_module_strict(self):

    """ Test that `Library` can load a module (by mod) in `strict` mode """

    with runtime.Library(operator, strict=True) as (library, _operator):
      assert hasattr(_operator, 'eq')



# -*- coding: utf-8 -*-

"""

  RPC tests
  ~~~~~~~~~

  tests canteen's server-side RPC layer.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

if __debug__:

  # RPC tests
  from . import test_base
  from . import test_exceptions

  # protocol tests
  from . import test_json
  from . import test_msgpack


  __all__ = ('test_base',
             'test_exceptions',
             'test_json',
             'test_msgpack')



# -*- coding: utf-8 -*-

"""

  core model adapters
  ~~~~~~~~~~~~~~~~~~~

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# stdlib
import json
import base64
import datetime
import collections

# mixin adapters
from .abstract import (KeyMixin,
                       EdgeMixin,
                       ModelMixin,
                       VertexMixin,
                       IndexedModelAdapter)


class AdaptedKey(KeyMixin):

  """ Provides bridged methods between `model.Key` and the Adapter API. """

  ## = Public Methods = ##
  def get(self, adapter=None):

    """ Retrieve a previously-constructed key from available persistence
        mechanisms. """

    if adapter:  # pragma: no cover
      return adapter._get(self)
    return self.__adapter__._get(self)

  def delete(self):

    """ Delete a previously-constructed key from available persistence
        mechanisms. """

    if self.__owner__:
      # if possible, delegate to owner model
      return self.__owner__.__adapter__._delete(self)
    return self.__class__.__adapter__._delete(self)

  def flatten(self, join=False):

    """ Flatten this Key into a basic structure suitable for transport or
        storage.

        :param join:
        :returns: """

    from canteen import model

    flat = []
    for element in (getattr(self, i) for i in reversed(self.__schema__)):
      if not isinstance(element, model.Key):
        flat.append(element)
      else:
        flat.append(element.flatten(join))

    flat = tuple(flat)
    if join:
      return self.__class__.__separator__.join([
        u'' if i is None else unicode(i) for i in (
          map(lambda x: x[0] if isinstance(x, tuple) else x, flat))]), flat
    return flat

  def urlsafe(self, joined=None):

    """ Generate an encoded version of this Key, suitable for use in URLs.

        :param joined:
        :returns: """

    if not joined: joined, flat = self.flatten(True)
    return base64.b64encode(joined)

  ## = Class Methods = ##
  @classmethod
  def from_raw(cls, encoded, **kwargs):

    """ Inflate a Key from a raw, internal representation.

        :param encoded:
        :param kwargs:
        :returns: """

    # if it's still a string, split by separator
    # (probably coming from a DB driver, `urlsafe` does this for us)
    encoded = (collections.deque(encoded.split(cls.__separator__))
               if isinstance(encoded, basestring) else (
                collections.deque(encoded)))

    key, keys = [], []
    if not (len(encoded) > len(cls.__schema__)):
      return cls(*encoded, **kwargs)

    last_key = encoded.popleft()  # we're dealing with ancestry here
    while len(encoded) > 2:
      # recursively decode, removing chunks as we go. extract argset by argset.
      last_key = cls(*(encoded.popleft() for i in (
                     xrange(0, len(cls.__schema__) - 1))),
                     parent=last_key or None,
                     _persisted=kwargs.get('_persisted', False))
    return cls(*encoded,
                parent=last_key or None,
                _persisted=kwargs.get('_persisted', False))

  @classmethod
  def from_urlsafe(cls, encoded, _persisted=False):

    """ Inflate a Key from a URL-encoded representation.

        :param encoded:
        :param _persisted:
        :returns: """

    return cls.from_raw(base64.b64decode(encoded), _persisted=_persisted)


class AdaptedModel(ModelMixin):

  """ Provides bridged methods between `model.Model` and the Adapter API. """

  ## = Public Class Methods = ##
  @classmethod
  def get(cls, key=None, name=None, adapter=None, **kwargs):

    """ Retrieve a persisted version of this model via the current model
        adapter.

        :param key:
        :param name:
        :param kwargs:

        :raises:
        :returns: """

    if not key and not name:
      raise ValueError('Must pass either a Key or'
                       ' key name into'
                       ' `%s.get`.' % cls.kind())  # pragma: no cover

    adapter = adapter or cls.__adapter__

    if name:
      # if we're passed a name, construct a key with the local kind
      return adapter._get(cls.__keyclass__(cls.kind(), name), **kwargs)
    if isinstance(key, basestring):
      # assume URL-encoded key, this is user-facing
      key = cls.__keyclass__.from_urlsafe(key)
    elif isinstance(key, (list, tuple)):
      key = cls.__keyclass__(*key)  # an ordered partslist is fine too
    return adapter._get(key, **kwargs)

  @classmethod
  def query(cls, *args, **kwargs):

    """ Start building a new `model.Query` object, if the underlying adapter
        implements `IndexedModelAdapter`.

        :param args:
        :param kwargs:

        :raises:
        :raises:
        :returns: """

    # we implement indexer operations
    if isinstance(cls.__adapter__, IndexedModelAdapter):
      from canteen.model import query

      filters, sorts = [], []
      for arg in args:
        if isinstance(arg, query.Filter):
          filters.append(arg)
        elif isinstance(arg, query.Sort):
          sorts.append(arg)
        else:
          raise RuntimeError('Cannot sort or filter based on'
                             ' arbitrary objects.'
                             ' Got: "%s".' % arg)  # pragma: no cover

      return query.Query(cls,
                         filters=filters,
                         sorts=sorts,
                         options=(
                          kwargs['options'] if 'options' in kwargs else (
                            query.QueryOptions(**kwargs))))

    else:  # pragma: no cover
      context = (cls.__adapter__.__class__.__name__, cls.kind())
      raise AttributeError("%s (currently selected for %s) does not"
                           " support indexing, and therefore can't"
                           " work with `model.Query` objects." % context)

  ## = Public Methods = ##
  def put(self, adapter=None, **kwargs):

    """ Persist this entity via the current model adapter.

        :param adapter:
        :param kwargs:
        :returns: """

    # allow adapter override
    if not adapter: adapter = self.__class__.__adapter__
    return adapter._put(self, **kwargs)

  def delete(self, adapter=None, **kwargs):

    """ Discard any primary or index-based data linked to this Key.

        :param adapter:
        :param kwargs:
        :returns: """

    # allow adapter override
    if not adapter: adapter = self.__class__.__adapter__
    return adapter._delete(self.__key__, **kwargs)


class AdaptedVertex(VertexMixin):

  """ Provides graph-oriented methods for ``Vertex`` objects. """

  __graph__ = __vertex__ = True  # mark as graph model and vertex

  def edges(self, tails=None, AND=None, OR=None, **kwargs):

    """ Retrieve edges for the current ``Vertex``.

        :param type:
        :param tails:
        :param heads:
        :param kwargs:

        :raises:
        :returns: """

    from canteen.model.query import EdgeFilter
    return self.query(**kwargs).filter(EdgeFilter(self.key, tails, **{
        'AND': AND, 'OR': OR}))

  def neighbors(self, tails=None, AND=None, OR=None, **kwargs):

    """ Retrieve neighbors (peered edges) for the current ``Vertex``.

        :param type:
        :param tails:
        :param kwargs:

        :raises:
        :returns: """

    from canteen.model.query import EdgeFilter
    return self.query(**kwargs).filter(EdgeFilter(self.key, tails, **{
        'AND': AND, 'OR': OR, 'type': EdgeFilter.NEIGHBORS}))


class AdaptedEdge(EdgeMixin):

  """ Provides graph-oriented methods for ``Edge objects``. """

  __graph__ = __edge__ = True  # mark as graph model and vertex


class DictMixin(ModelMixin):

  """ Provides `to_dict`-type methods for first-class Model API classes. """

  def update(self, mapping=None, **kwargs):

    """ Update properties on this model via a merged dict of mapping + kwargs.

        :param mapping:
        :param kwargs:
        :returns: """

    mapping = mapping or {}
    if kwargs: mapping.update(kwargs)
    map(lambda x: setattr(self, x[0], x[1]), mapping.items())
    return self

  def to_dict(self, exclude=tuple(), include=tuple(),
              filter=None, map=None, _all=False,
              filter_fn=filter, map_fn=map,
              convert_keys=True, convert_models=True, convert_datetime=True):

    """ Export this Entity as a dictionary, excluding/including/ filtering/
        mapping as we go.

        :param exclude:
        :param include:
        :param filter:
        :param map:
        :param _all:
        :param filter_fn:
        :param map_fn:
        :param convert_keys:
        :param convert_models:

        :raises:
        :returns: """

    from canteen import model

    dictionary = {}  # return dictionary
    _default_include = False  # flag for including properties unset

    # explicit mode implies returning all properties raw
    if not _all: _all = self.__explicit__

    if not include:
      include = self.__lookup__  # default include list is model properties
      _default_include = True  # mark flag that we used the default

    if not map:
      map = lambda x: x  # substitute no map with a passthrough
      _default_map = True

    # substitute no filter with a passthrough
    if not filter: filter = lambda x: True


    # freeze our comparison sets
    exclude, include = frozenset(exclude), frozenset(include)

    for name in self.__lookup__:

      # run map fn over (name, value)
      _property_descriptor = self.__class__.__dict__[name]


      # pull with property default
      _prop_default = self.__class__.__dict__[name].default
      name, value = map((name, self._get_value(name, default=_prop_default)))

      # run filter fn over (name, vlaue)
      filtered = filter((name, value))
      if not filtered: continue

      # filter out via exclude/include
      if name in exclude:
        continue
      if not _default_include:
        if name not in include: continue

      if value is _property_descriptor.sentinel:  # property is unset
        # if it matches an item in a custom include list,
        # and/or we don't want all properties...
        if not _all and not ((not _default_include) and name in include):
          continue  # skip if all properties not requested
        else:
          if not self.__explicit__:  # None == sentinel in implicit mode
            dictionary[name] = None
            continue

      _bundle = []

      # validate inner value which may be repeated
      for _val in ((value,) if not isinstance(value, (list, tuple)) else value):

        if isinstance(_val, (model.Model, model.Key)):
          if _property_descriptor.options.get('embedded'):
            if isinstance(_val, model.Key):  # pragma: no cover
              raise RuntimeError('Cannot reference embedded submodel'
                                 ' by key "%s".' % repr(_val))
            _bundle.append(_val.to_dict() if convert_models else _val)
          else:
            if isinstance(_val, model.Model) and not (
                  _val.key):  # pragma: no cover
              raise RuntimeError('Cannot reference non-embedded submodel'
                                 ' "%s" with empty key.' % repr(_val))

            # no embedded status means it should be left as-is
            elif _property_descriptor.options.get('embedded') is None:

              # @TODO(sgammon): decide sensible logic here

              # if mapping is an explicit key, conver it to a string
              if _property_descriptor.basetype is model.Key and (
                    convert_keys):
                _bundle.append(_val.urlsafe())
              else:
                _bundle.append(_val)

            else:  # pragma: no cover
              if isinstance(value, model.Key) and convert_keys:
                _bundle.append(_val.urlsafe())
              elif isinstance(value, model.Model) and convert_models:
                _bundle.append(_val.to_dict())
              else:
                _bundle.append(_val)
        elif isinstance(value, (datetime.date, datetime.datetime)):
          _bundle.append((
            _val if not convert_datetime else _val.isoformat()))
        else:
          _bundle.append(_val)

      if _property_descriptor.repeated:
        dictionary[name] = (
          tuple(_bundle) if isinstance(value, tuple) else _bundle)
      else:
        dictionary[name] = _bundle.pop()
    return dictionary

  @classmethod
  def to_dict_schema(cls):

    """ Convert a model or entity's schema to a dictionary, where keys=>values
        map to properties=>descriptors.

        :returns: """

    schema = {}
    for name in cls.__lookup__:
      schema[name] = getattr(cls, name)
    return schema


class JSONMixin(KeyMixin, ModelMixin):

  """ Provides JSON serialization/deserialization support to `model.Model` and
      `model.Key`. """

  def to_json(self, *args, **kwargs):

    """ Convert an entity to a JSON structure, where keys=>values map to
        properties=>values.

        :param args:
        :param kwargs:
        :returns: """

    return json.dumps(self.to_dict(*args, **kwargs))

  @classmethod
  def from_json(cls, encoded):

    """ Inflate a JSON string into an entity. Expects a dictionary of
        properties=>values.

        :param encoded:
        :returns: """

    return cls(**json.loads(encoded))

  @classmethod
  def to_json_schema(cls, *args, **kwargs):  # pragma: no cover

    """ Convert a model or entity's schema to a dictionary, where keys=>values
        map to JSON Schema representing properties=>descriptors.

        :param args:
        :param kwargs:

        :raises:
        :returns: """

    raise NotImplementedError()  # @TODO: JSON schema support


# msgpack support
try:
  import msgpack
except ImportError as e:  # pragma: no cover
  pass  # no `msgpack` support :(

else:


  class MsgpackMixin(KeyMixin, ModelMixin):

    """ Provides Msgpack serialization/deserialization support to `model.Model`
        and `model.Key`. """

    def to_msgpack(self, *args, **kwargs):

      """ Convert an entity to a Msgpack structure, where keys=>values map to
          properties=>values.

          :param args:
          :param kwargs:

          :returns: """

      return msgpack.dumps(self.to_dict(*args, **kwargs))

    @classmethod
    def from_msgpack(cls, encoded):

      """ Inflate a msgpack payload into an entity. Expects a dictionary of
          properties=>values.

          :param encoded:
          :returns: """

      return cls(**msgpack.unpackb(encoded))

    @classmethod
    def to_msgpack_schema(cls, *args, **kwargs):  # pragma: no cover

      """ Convert a model or entity's schema to a dictionary, where keys=>values
          map to internal symbols representing properties=>descriptors.

          :param args:
          :param kwargs:

          :raises:
          :returns: """

      raise NotImplementedError()  # @TODO: msgpack schema support?



# -*- coding: utf-8 -*-

"""

  in-memory model adapter
  ~~~~~~~~~~~~~~~~~~~~~~~

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# stdlib
import time
import json
import base64
import datetime
import itertools
import collections

# adapter API
from .abstract import DirectedGraphAdapter


## Globals
_init = False
_graph = {}
_metadata = {}
_datastore = {}


## Constants
_sorted_types = (int,
                 long,
                 float,
                 datetime.date,
                 datetime.datetime)


## Utils
_to_timestamp = lambda dt: int(time.mktime(dt.timetuple()))


class InMemoryAdapter(DirectedGraphAdapter):

  """ Adapt model classes to RAM with a simple adapter. Mainly meant as a
      reference ``DirectedGraphAdapter`` implementation. Supports querying
      and graph storage. """

  # key encoding
  _key_encoder = base64.b64encode

  # data compression / encoding
  _data_encoder = json.dumps
  _data_compressor = None

  is_supported = classmethod(lambda cls: True)  # always supported

  @classmethod
  def acquire(cls, name, bases, properties):

    """ Perform first initialization.

        :param name: Target :py:class:`Model` subtype class name.
        :param bases: Base classes for target subtype.
        :param properties: Class-level property mappings for target subtype.

        :returns: Prepared :py:class:`InMemoryAdapter` object for target
          ``(name, bases, properties)`` combo. """

    global _init, _graph, _metadata

    # perform first init, if it hasn't been done
    if not _init:
      _init, _metadata, _graph = True, {

        'ops': {  # holds count of performed operations
          'get': 0,  # track # of entity get() operations
          'put': 0,  # track # of entity put() operations
          'delete': 0},  # track # of entity delete() operations

        'keys': set(),  # holds set of all known keys
        'kinds': {},  # holds current count and ID increment
        'global': {  # holds global metadata, like entity count
          'entity_count': 0,  # holds global count of all entities
          'node_count': 0,  # holds count of known nodes
          'edge_count': 0},  # holds count of known edges

        cls._key_prefix: set([]),  # full, simple indexed set of all keys
        cls._kind_prefix: {},  # maps keys to their kinds
        cls._group_prefix: {},  # maps keys to their entity groups
        cls._index_prefix: {},  # maps property values to keys
        cls._reverse_prefix: {}  # maps keys to indexes they are present in

      }, {

        # holds nodes (structurally)
        'nodes': collections.defaultdict(lambda: set()),

        # holds edges (structurally)
        'edges': {

          # holds directed edges
          'directed': {
            'in': collections.defaultdict(lambda: set()),
            'out': collections.defaultdict(lambda: set())},

          # holds undirected edges
          'undirected': collections.defaultdict(lambda: set())},

        # holds neighbor indexes
        'neighbors': {

          # holds directed edge neighbor indexes
          'directed': {
            'in': collections.defaultdict(lambda: set()),
            'out': collections.defaultdict(lambda: set())},

          # holds undirected neighbor indexes
          'undirected': collections.defaultdict(lambda: set())}}

    # pass up the chain to create a singleton
    return super(InMemoryAdapter, cls).acquire(name, bases, properties)

  @classmethod
  def get(cls, key, **kwargs):

    """ Retrieve an entity by Key from Python RAM.

        :param key: Target :py:class:`model.Key` object at which data should be
          fetched, if it is available.

        :param kwargs: Implementation-specific flags/kwargs passed to the
          underlying adapter from the application.

        :returns: Entity at ``key``, if any, or ``None`` if no entity could
          be found at ``key``. """

    global _metadata

    # key format: tuple(<str encoded key>, <tuple flattened key>)
    encoded, flattened = key

    # pull from in-memory backend
    entity = _datastore.get(flattened)
    if entity is None: return  # not found

    _metadata['ops']['get'] += 1

    return entity  # construct + inflate entity

  @classmethod
  def put(cls, key, entity, model, **kwargs):

    """ Persist an entity to storage in Python RAM.

        :param key: Target :py:class:`model.Key` object at which data should be
          persisted.

        :param entity: Entity object to store against ``key`` in RAM.

        :param model: :py:class:`model.Model` subtype kind for ``entity``. Used
          for property/policy reference.

        :param kwargs: Implementation-specific flags/kwargs to the underlying
          adapter from the application.

        :returns: ``key`` at which ``entity`` was stored in underlying storage,
          including any elements of ``key`` which required population from
           things like ID provisioning. """

    global _graph, _metadata, _datastore

    # encode key and flatten
    encoded, flattened = key
    target = flattened

    # perform validation
    with entity:

      if entity.key.kind not in _metadata['kinds']:  # pragma: no cover
        _metadata['kinds'][entity.key.kind] = {
          'id_pointer': 0,  # keep current key ID pointer
          'entity_count': 0,  # keep count of seen entities for each kind
          'keys': set()}  # keep set of all keys for kind

      # update counts
      _metadata['ops']['put'] = _metadata['ops'].get('put', 0) + 1

      _metadata['global']['entity_count'] = (
        _metadata['global'].get('entity_count', 0) + 1)

      kinded_entity_count = (
        _metadata['kinds'][entity.key.kind].get('entity_count', 0))

      _metadata['kinds'][entity.key.kind]['entity_count'] = (
        kinded_entity_count + 1)

      # add to keys for kind
      kinded_keys = _metadata['kinds'][entity.key.kind].get('keys', set())
      kinded_keys.add(target)
      _metadata['kinds'][entity.key.kind]['keys'] = kinded_keys

      # add to main keys index
      _metadata['keys'].add(target)

      # save to datastore
      _datastore[target] = entity

      # store vertexes separately
      if getattr(model, '__vertex__', False):
        _graph['nodes'][entity.key.kind].add(target)

      # store edges separately
      elif getattr(model, '__edge__', False):

        # directed edges
        if model.__spec__.directed:

          left = _graph['neighbors']['directed']['out'][entity['source']]
          left_e = _graph['edges']['directed']['out'][entity['source']]

          for _edge_target in entity['target']:
            right = _graph['neighbors']['directed']['in'][_edge_target]
            right_e = _graph['edges']['directed']['in'][_edge_target]

            # index neighbors to each other
            right.add(entity['source'])
            right_e.add(entity.key)

            # index edges
            left.add(_edge_target)
            left_e.add(entity.key)

        # undirected edges
        else:
          for origin in entity['peers']:
            for target in entity['peers']:
              if origin == target: continue
              _graph['neighbors']['undirected'][origin].add(target)
              _graph['edges']['undirected'][origin].add(entity.key)
              _graph['edges']['undirected'][target].add(entity.key)

    return entity.key

  @classmethod
  def delete(cls, key, **kwargs):

    """ Delete an entity by Key from memory.

        :param key: Target :py:class:`model.Key` object at which data should be
          deleted in underlying RAM storage.

        :param kwargs: Implementation-specific flags/kwargs to the underlying
          adapter from the application.

        :returns: ``True`` if the entity at ``key`` was found and deleted,
          ``False`` if the entity could not be found for deletion. """

    global _metadata, _datastore

    # extract key
    if not isinstance(key, tuple):  # pragma: no cover
      encoded, flattened = key.flatten(True)
    else:
      encoded, flattened = key

    # extract key parts
    parent, kind, _id = flattened

    # if we have the key...
    if flattened in _metadata[cls._key_prefix]:
      try:
        del _datastore[flattened]  # delete from datastore

      except KeyError:  # pragma: no cover
        _metadata[cls._key_prefix].remove(flattened)
        return False  # untrimmed key

      else:
        # update meta
        _metadata[cls._key_prefix].remove(flattened)
        _metadata['ops']['delete'] = (
          _metadata['ops'].get('delete', 0) + 1)

        _metadata['global']['entity_count'] = (
          _metadata['global'].get('entity_count', 1) - 1)

        _metadata['kinds'][kind]['entity_count'] = (
          _metadata['kinds'][kind].get('entity_count', 1) - 1)

      return True
    return False

  @classmethod
  def allocate_ids(cls, key_class, kind, count=1, **kwargs):

    """ Allocate new Key IDs up to `count`.

        :param key_class: :py:class:`model.Key` subtype for which we are
          allocating integer IDs from underlying RAM storage.

        :param kind: ``str`` or ``unicode`` kind name for which we are
          allocating integer IDs from underlying RAM storage. Used to resolve
          an appropriate ``id_pointer`` index.

        :param count: Number of unique key IDs we would like to provision.
          Defaults to ``1``.

        :param kwargs: Implementation-specific flags/kwargs to the underlying
          adapter from the application.

        :raises StopIteration: When all the requested IDs have been allocated
          and yielded to the application.

        :returns: If a ``count`` is given that is ``> 1``, will return a
          generator function that will ``yield`` IDs up to ``count`` one at a
          time. Otherwise, returns the newly-provisioned ID integer directly,
          making the return type either ``function`` (if ``count`` is greater
          than one) or ``int``/``long``. """

    global _metadata

    # resolve kind meta and increment pointer
    kind_blob = _metadata['kinds'].get(kind, {})
    current = kind_blob.get('id_pointer', 0)
    pointer = kind_blob['id_pointer'] = (current + count)

    # update kind blob
    _metadata['kinds'][kind] = kind_blob

    # return IDs
    if count > 1:
      def _generate_id_range():
        for x in xrange(current, pointer):
          yield x
        raise StopIteration()
      return _generate_id_range
    return pointer

  @classmethod
  def write_indexes(cls, writes, _graph, execute=True):

    """ Write a set of generated indexes via `generate_indexes`.

        :param writes: Index writes to commit to the ``_metadata`` ``dict``.

        :param _graph: Graph writes to commit. Ignored in this adapter as those
          are handled in ``put`` instead.

        :param execute: ``bool`` flag, whether to execute and actually commit
          the writes planned.

        :raises RuntimeError: If an invalid bundle of index writes is provided.

        :returns: Writes committed to the ``_metadata`` ``dict``, for
          inspection. """

    global _metadata

    _write = {} if not execute else _metadata

    # extract indexes
    target, meta, properties = writes

    # write indexes one-by-one, generating reverse entries as we go
    for serializer, write in itertools.chain(
      ((None, _m) for _m in meta), (bundle for bundle in properties)):

      # filter out strings, convert to 1-tuples
      if isinstance(write, basestring):  # pragma: no cover
        write = (write,)

      if len(write) > 3:  # hashed/mapped index

        # extract write, inflate
        index, path, value = write[0], write[1:-1], write[-1]

        if isinstance(value, dict):  # pragma: no cover
          continue  # cannot index dictionaries

        # convert into a compound sorted index
        if isinstance(value, _sorted_types):

          if isinstance(value, datetime.datetime):
            value = _to_timestamp(value)

          write = (index, path, (value, target))

        else:
          # init index hash (mostly covers custom indexes)
          if index not in _write:  # pragma: no cover
            _write[index] = {(path, value): set()}

          elif (path, value) not in _write[index]:
            _write[index][(path, value)] = set()

          # write key to index
          _write[index][(path, value)].add(target)

          # add reverse index
          if target not in _write[cls._reverse_prefix]:  # pragma: no cover
            _write[cls._reverse_prefix][target] = set()
          _write[cls._reverse_prefix][target].add((index, path, value))

          continue

      if len(write) == 3:  # pragma: no cover
        # @TODO(sgammon): Do we need this?

        # simple map index

        # extract write, inflate
        index, dimension, value = write

        # init index hash
        if index not in _write:
          _write[index] = {dimension: {value}}

        elif dimension not in _write[index]:
          _write[index][dimension] = {value}

        else:
          # everything is there, map the value
          _write[index][dimension].add(value)

        # add sorted mark, if necessary
        if isinstance(value, tuple) and isinstance(value[0], _sorted_types):
          _mark = (dimension, '__sorted__')
          if _mark not in _write[index]:
            _write[index][_mark] = {}
          _write[index][_mark][target] = value

        # add reverse index
        if target not in _write[cls._reverse_prefix]:
          _write[cls._reverse_prefix][target] = set()
        _write[cls._reverse_prefix][target].add((index, dimension))
        continue

      elif len(write) == 2:  # simple set index

        # extract write, inflate
        index, value = write

        # init index hash
        if index not in _write:  # pragma: no cover
          _write[index] = {value: set()}

        # init value set
        elif value not in _write[index]:
          _write[index][value] = set()

        # only provision if value and index are different
        if index != value:

          # add flattened key
          _write[index][value].add(target)

        # add reverse index
        if target not in _write[cls._reverse_prefix]:
          _write[cls._reverse_prefix][target] = set()
        _write[cls._reverse_prefix][target].add(index)
        continue

      elif len(write) == 1:  # simple key mapping

        # extract singular index
        index = write[0]

        # special case: key index
        if index == cls._key_prefix:
          _write[index].add(target)
          continue

        # provision index
        if index not in _write:  # pragma: no cover
          _write[index] = {}

        # add value to index
        _write[index][target] = set()  # provision with a one-index entry

        # add reverse index
        if target not in _write[cls._reverse_prefix]:  # pragma: no cover
          _write[cls._reverse_prefix][target] = set()
        _write[cls._reverse_prefix][target].add((index,))
        continue

      else:  # pragma: no cover
        raise RuntimeError("Index mapping tuples must have at least 2 entries,"
                           " for a simple set index, or more for"
                           " a hashed index.")

    return _write

  @classmethod
  def clean_indexes(cls, writes, **kwargs):

    """ Clean indexes for a key that is due to be deleted.

        :param writes: Index writes that would be committed if the ``key`` was
          being written. Used to provide a directory of index items to clean
          from underlying RAM.

        :param kwargs: Implementation-specific flags/kwargs to the underlying
          adapter from the application.

        :returns: ``set`` instance containing index keys that were cleaned. """

    global _metadata

    target, meta, graph = writes  # extract indexes

    # pull reverse indexes
    reverse = _metadata[cls._reverse_prefix].get(target, set())

    _cleaned = set()  # clear reverse indexes
    if len(reverse) or len(meta):
      for i in reverse | set(meta):

        # convert to tuple to be consistent
        if not isinstance(i, tuple):
          i = (i,)

        # check cleanlist
        if i in _cleaned:  # pragma: no cover
          continue  # we've already cleaned this directive
        else:
          _cleaned.add(i)

        if len(i) == 3:  # hashed index

          # extract write, clean
          index, path, value = i

          if isinstance(path, tuple):
            if index in _metadata and (path, value) in _metadata[index]:
              _metadata[index][(path, value)].remove(target)

              # if there's no keys left in the index, trim it
              if len(_metadata[index][(path, value)]) == 0:
                del _metadata[index][(path, value)]

            continue

          # (mostly covers custom indexes)
          if isinstance(path, basestring):  # pragma: no cover
            if index in _metadata and path in _metadata[index]:
              _metadata[index][path].remove(target)

              # if there's no keys left in the entry, trim it
              if len(_metadata[index][path]) == 0:
                del _metadata[index][path]

            continue

        elif len(i) == 2:  # simple set index
          # extract write, clean
          index, value = i

          if index in _metadata and value in _metadata[index]:

            # check sorted-ness
            svalue = (value, '__sorted__')
            if svalue in _metadata[index]:
              sorted_entry = _metadata[index][svalue].get(target)
              if sorted_entry:
                _metadata[index][value].remove(sorted_entry)

            else:
              # remove from set at item in mapping
              _metadata[index][value].remove(target)

            # if there's no keys left in the index, trim it
            if len(_metadata[index][value]) == 0:  # pragma: no cover
              del _metadata[index][value]

          continue

        elif len(i) == 1:  # simple key mapping
          if i[0] == '__key__':
            continue  # skip keys, that's done by `delete()`
          if target in _metadata[i[0]]:
            del _metadata[i[0]][target]

    if target in _metadata[cls._reverse_prefix]:
      # last step: remove reverse index for key
      del _metadata[cls._reverse_prefix][target]

    return _cleaned

  @classmethod
  def encode_key(cls, joined, flattened):

    """ Encode a :py:class:`model.Key` for storage. For ``InMemoryAdapter``,
        simply use the flattened tuple as an internal representation of the
        key.

        :param key: Target :py:class:`model.Key` to encode.

        :param joined: Joined/stringified key.

        :param flattened: Flattened ``tuple`` (raw) key.

        :returns: The encoded :py:class:`model.Key`, or ``False`` to
              yield to the default encoder. """

    return flattened

  @classmethod
  def execute_query(cls, kind, spec, options, **kwargs):  # pragma: no cover

    """ Execute a query across one (or multiple) indexed properties. Collapses
        a symbolic :py:class:`canteen.model.query.Query` object and attempts to
        properly satisfy any ``Filter``/``Sort`` objects attached.

        :param kind: :py:class:`model.Model` subtype class that we're querying
          for. Used for resolving policy/index names/storage names.

        :param spec: Tuple of ``(filters, sorts)`` to apply for this ``Query``
          execution run.

        :param options: :py:class:`canteen.model.query.QueryOptions` instance,
          which specifies query options like a result ``offset`` or ``limit``.

        :param kwargs: Implementation-specific flags/kwargs to the underlying
          adapter from the application.

        :raises RuntimeError: In the event of an unsatisfiable query or other
          unrecoverable runtime error.

        :returns: Results matching ``spec`` for ``kind`` according to
          ``options``, or an empty ``list`` if no results could be found. """

    from canteen import model
    from canteen.model import query

    # extract spec
    filters, sorts = spec

    # calculate ancestry parent
    ancestry_parent = None
    if isinstance(options.ancestor, basestring):
      ancestry_parent = model.Key.from_urlsafe(options.ancestor)
    elif isinstance(options.ancestor, model.Key):
      ancestry_parent = options.ancestor
    elif isinstance(options.ancestor, model.Model):
      ancestry_parent = options.ancestor.key

    # prepare workspace
    _data_frame, _q_init = set(), False
    _special_indexes, _sorted_indexes, _unsorted_indexes, _inmemory_filters = (
        [], [], [], [])

    _index_groups = (_special_indexes, _sorted_indexes, _unsorted_indexes)

    ## apply ancestry first
    if ancestry_parent:
      _ekey = cls.encode_key(*ancestry_parent.flatten(True))
      _group_index = _metadata[cls._group_prefix].get(_ekey)
      if _group_index: _special_indexes.append((False, (
          query.KeyFilter(_ekey, type=query.KeyFilter.ANCESTOR), _group_index)))

    ## apply filters
    if filters or ancestry_parent:

      for _f in filters:

        if isinstance(_f.value, model.Model._PropertyValue):
          _filter_val = _f.value.data
        else:
          _filter_val = _f.value

        if isinstance(_f.value.data, _sorted_types):

          # convert timestamps
          if isinstance(_f.value.data, (datetime.datetime, datetime.date)):
            _filter_val = _to_timestamp(_f.value.data)

          # valued index
          _index_key = (kind.__name__, _f.target.name)
          if _index_key in _metadata[cls._index_prefix]:
            _sorted_indexes.append((True, (
              _f.target.name,
              _f.operator,
              _filter_val,
              _metadata[cls._index_prefix][_index_key])))

        else:

          if isinstance(_f, query.EdgeFilter):

            _graph_base = 'edges' if _f.kind is _f.EDGES else 'neighbors'

            if _f.tails is None:  # undirected query
              _target_edge_index = (
                _graph[_graph_base]['undirected'].get(_filter_val, set()))

            else:  # directed queries

              _direction = 'out' if _f.tails else 'in'
              _target_edge_index = (
                _graph[_graph_base]['directed'][_direction].get(*(
                    _filter_val, set())))

            _unsorted_indexes.append((False, (_f, _target_edge_index)))

          else:
            # devalued index
            _index_key = ((kind.__name__, _f.target.name), _f.value.data)
            if _index_key in _metadata[cls._index_prefix]:
              _unsorted_indexes.append((
                False, (_f, _metadata[cls._index_prefix][_index_key])))

      for group in _index_groups:
        for is_sorted, directive in group:

          # sorted indexes
          if is_sorted:
            target, operator, value, index = directive
            high_bound = (value if (operator in (
              query.LESS_THAN, query.LESS_THAN_EQUAL_TO)) else None)
            low_bound = (value if (operator in (
              query.GREATER_THAN, query.GREATER_THAN_EQUAL_TO)) else None)

            if low_bound and not high_bound:
              evaluate = lambda bundle: (low_bound < bundle[0])

            elif low_bound and high_bound:
              evaluate = lambda bundle: (low_bound < bundle[0] < high_bound)

            elif high_bound:  # high-bound only
              evaluate = lambda bundle: (bundle[0] < high_bound)

            else:  # invalid filter
              raise RuntimeError('Invalid sorted filter'
                                 ' operation: "%s".' % operator)

            if not _q_init:  # no frame yet, initialize
              _q_init = True
              _data_frame = (
                set((value for _, value in filter(evaluate, index))))
            else:  # otherwise, filter
              _data_frame &= (
                set((value for _, value in filter(evaluate, index))))

          # unsorted indexes
          else:
            _fblock, _target = directive

            if _fblock.operator in frozenset((query.EQUALS,
                                              query.CONTAINS,
                                              query.KEY_KIND,
                                              query.KEY_ANCESTOR)):

              if not _q_init:  # no frame yet, initialize
                _q_init = True
                _data_frame = _target
              else:
                # filter against selected target
                _data_frame &= _target
            elif _fblock.operator is query.NOT_EQUALS:
              # inequality filters are deferred
              _inmemory_filters.append(_fblock)
            else:  # pragma: no cover
              raise RuntimeError('Invalid query operator "%s" encountered'
                                  ' during execution.' % _fblock.operator)

    elif not filters and not ancestry_parent:
      # no filters - working with _all_ models of a kind as base
      _data_frame = _metadata[cls._kind_prefix].get(kind.__name__)

    ## inflate results (keys only)
    if options.keys_only and not _inmemory_filters:
      _keyify = lambda k: (
        model.Key.from_urlsafe(k, _persisted=True) if not (
          isinstance(k, model.Key)) else k)

      return (_keyify(k) for k in _data_frame)

    result_entities = []

    if not _data_frame and _inmemory_filters:
      # corner case: no initial data frame but inmemory filters
      #  force-initialize data frame with kind index, or if we're unlucky
      #  enough to be doing a kindless query, the index of all available
      #  keys.

      if kind: _data_frame = _metadata['kinds'].get(kind.__name__, {'keys': set()})['keys']
      else: _data_frame = _metadata['keys']


    for n, (key, entity) in enumerate(((k, _datastore.get(k)) for k in _data_frame)):

      # @TODO(sgammon) log ghosts?
      if not entity: continue  # skip missing entities

      # stop if options say so, skip if the options say so
      if options.limit and n >= options.limit: break
      if options.offset and n <= options.offset: continue
      # @TODO(sgammon): make sure we don't get any off-by-ones

      # collapse in-memory filters
      for _inner_f in _inmemory_filters:
        if not _inner_f(entity):
          break
      else:
        result_entities.append(entity)

    if not result_entities: return result_entities  # no need to sort, obvs

    ## apply sorts
    if sorts:

      def do_sort(sort, results):

        """ Apply a ``sort`` operation to a set of query ``results``.

            :param sort: :py:class:`canteen.model.query.Sort` instance
              describing the sort operator to be applied and the target
              ``prop`` to apply it to.

            :param results: ``list`` of datastore ``results`` that have been
             filtered and retrieved from underlying storage.

            :returns: Query results passed in as ``results``, but sorted
              according to ``sort``. """

        _sort_i, _sort_frame, _sort_values = set(), [], {}

        # build value index and map to keys
        for result in results:
          val = getattr(result, sort.target.name, None)
          if val is not None:
            if val not in _sort_values:
              _sort_i.add(val)  # add to known values
              _sort_values[val] = set()  # provision value -> obj map
            _sort_values[val].add(result)  # add to map

        _rvs, _fwd = lambda d: reversed(sorted(d)), lambda d: sorted(d)

        if sort.target.basetype in (basestring, str, unicode):
          sorter = (_rvs if (sort.operator is query.ASCENDING) else _fwd)

        else:
          sorter = (_fwd if (sort.operator is query.ASCENDING) else _rvs)

        # choose iterator and start sorting
        for value in sorter(_sort_i):

          # quick optimization: no need to subgroup, just one sort
          if len(sorts) == 1:
            for _valued_result in _sort_values[value]:
              _sort_frame.append(_valued_result)

          # aww, we have to subgroup and subsort
          if len(sorts) > 1:
            # add sorted groups of values
            _sort_frame.append((value, _sort_values[value]))
        return _sort_frame

      if len(sorts) == 1:
        return do_sort(sorts[0], result_entities)

      if len(sorts) > 1:

        # sort atop the result of the last sort
        _sort_base = []
        for sort in sorts:
          _sort_base = do_sort(*(
            sort, (_sort_base and result_entities) or _sort_base))
        return _sort_base
    return result_entities



# -*- coding: utf-8 -*-

"""

  setup tests
  ~~~~~~~~~~~

  makes sure canteen's setup routine works properly.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# setup
from canteen import setup
